/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <climits>  

using namespace std;

#include "seqcode.hh"
#include "substmod.hh"
#include "indel.hh"
#include "treespec.hh"
#include "data.hh"


//+++++++++++++++++++++++++++++++++++++++++++
//             class dataset                +
//+++++++++++++++++++++++++++++++++++++++++++


dataset::dataset()
{
}

dataset::dataset(ifstream& datei,int seqformat)
{
  switch(seqformat) {
    case 0: //FASTA
      fromfasta(datei);
      break;
    case 1: //PHYLIP
      fromphylip(datei);  
      break;
    default:
      cerr << "Format " << seqformat << " not supported\n";
      exit(EXIT_FAILURE);
  }  
  datei.close();
  
  guesstype();
}

dataset::dataset(baum *b, alphsize alf)
{  
   vector<int> ali_to_node;
   vector<int> node_to_ali;
   typ=alf;
   push_mult(b,ali_to_node,node_to_ali,Identif,Seqs);
}

dataset::~dataset() {
}

void dataset::fromfasta(ifstream& datei)
{
      unsigned int i, flag;
      bool flag2;
      char c;

      i = 0;
      flag = 2;
      flag2=TRUE;
      while(datei.get(c))
      {
         switch(c) {
           case '>' :
             flag = 1;
	     flag2=TRUE;
	     i++;
	     Identif.resize(i);
	     Seqs.resize(i);
	     break;
           case '\n' :
	     if(flag==2) {
	        cerr << "The sequence file is not in FASTA format!\n";
		exit(EXIT_FAILURE);
	     }
             flag = 0;
	     break;
           default :
	     if(flag==2) {
	        cerr << "The sequence file is not in FASTA format!\n";
		exit(EXIT_FAILURE);
	     }
             if(isprint(c) && !isspace(c))
	     {
	       if(flag==1 && flag2) {
	           Identif[i-1].push_back(c); // Name nur bis zum ersten Leerzeichen
	       } else if(flag==0)
	       {
	         if(c!='-')
	           Seqs[i-1].push_back(toupper(c));
		 else
		   Seqs[i-1].push_back('_');
	       }
	     } else {
	       if(flag==1 && Identif[i-1].size()>0) // leading blanks erlaubt
	          flag2=FALSE;
             }
         }
      }
      if(Identif.size()==0) {
         cerr << "There is no data!\n";
	 exit(EXIT_FAILURE);
      }
      for(i=0;i<Identif.size();++i)
          if(Identif[i].size()==0) {
	     cerr << "There are sequences without names!\n";
	     exit(EXIT_FAILURE);
	  }
}

void dataset::fromphylip(ifstream& datei)
{
      unsigned int Maxspc, Maxseqc;
      unsigned int i, j, k, leerZ, Zeilen, Idcount, flag;
      vector<unsigned int> Position(1);
      string firstline="";
      char c;


      while(datei.get(c) && c!='\n')
          firstline+=c;
      if(sscanf(firstline.c_str(),"%u %u",&Maxspc,&Maxseqc)!=2) {
         cerr << "The sequence file is not in PHYLIP format\n";
	 exit(EXIT_FAILURE);
      }
      Identif.resize(Maxspc);
      for(i=0;i<Identif.size();i++)
      {
        Identif[i].resize(10);
        for(j=0;j<Identif[i].size();j++)
          Identif[i][j]='\0';
      }
      Seqs.resize(Maxspc);
      for(i=0;i<Seqs.size();i++)
      {
        Seqs[i].resize(Maxseqc);
        for(j=0;j<Seqs[i].size();j++)
          Seqs[i][j]='\0';
      }
      Position.resize(Maxspc);
      for(i=0;i<Position.size();i++)
        Position[i]=0;
      leerZ = 0;
      Zeilen = 0;
      Idcount = 0;
      flag = 0;
  
      do
      {
        if(c=='\n')
        {
          if(flag==0)
            leerZ++;
          Zeilen++;
          Idcount = 0;
          flag = 0;
          if((Zeilen-leerZ)%Seqs.size()==0)
          {
            //Kontrolle ob Zeilen gleich lang
            for(i=0;i<Position.size()-1;i++)
            {
              if(Position[i]!=Position[i+1])
              {
                cerr << "Something's wrong\n"
                     << "Position[" << i << "]="
	             << Position[i] << " != Position["
	             << i+1 << "]=" << Position[i+1] << "\n";
                exit(1);
              }
	    }
          }
        }
        else
        {
          if(isprint(c) && !isspace(c))
          {
            flag = 1;
	    if(Zeilen==0)
	    {
	      cerr << "Please remove text in the first line\n";
	      exit(1);
	    }
          }
          if(flag==1)
          {
            Idcount++;
            if(Idcount<=Identif[0].size() && (Zeilen-leerZ)<Identif.size())
              Identif[Zeilen-leerZ][Idcount-1]=c;
            else
            {
              if(Position[(Zeilen-leerZ)%Seqs.size()]<Seqs[0].size())
	      {
                if(isalpha(c) || c=='-'  || c=='_' || c=='*' || c=='?')
	        {
		  if(c!='-')
	            Seqs[(Zeilen-leerZ)%Seqs.size()][Position[(Zeilen-leerZ)%Seqs.size()]]=toupper(c);
		  else
		    Seqs[(Zeilen-leerZ)%Seqs.size()][Position[(Zeilen-leerZ)%Seqs.size()]]='_';
	          Position[(Zeilen-leerZ)%Seqs.size()]++;
	        }
	        else
	        {
	          if(c=='.')
	          {
	            if((Zeilen-leerZ)%Seqs.size()>0)
	            {
	              Seqs[(Zeilen-leerZ)%Seqs.size()][Position[(Zeilen-leerZ)%Seqs.size()]]=Seqs[0][Position[(Zeilen-leerZ)%Seqs.size()]];
	              Position[(Zeilen-leerZ)%Seqs.size()]++;
	            }
	            else
	            {
	              cerr << "Error in Phylip file: dot in the first sequence\n";
		      exit(1);
	            }
	          }
	        }
	      }
            }
          }
        }
  
      } while(datei.get(c) && Position[Position.size()-1]<Seqs[0].size());
  
      //Kontrolle
      for(i=0;i<Position.size()-1;i++)
      {
        if(Position[i]!=Position[i+1])
        {
          cerr << "Something's wrong\n"
               << "Position[" << i << "]="
	       << Position[i] << " != Position["
	       << i+1 << "]=" << Position[i+1] << "\n";
          exit(1);
        }
        if(Position[i]!=Seqs[0].size() || Position[i+1]!=Seqs[0].size())
        {
          cerr << "Something's wrong.\n"
               << "Not enough nucleotides in the sequence\n"
	       << i+1 << " or " << i+2 << "\n";
          exit(1);      
        }
       }
      //leading blanks in Identif
      for(i=0;i<Identif.size();i++)
      {
        flag = 0;
        for(j=1;j<=Identif[0].size();j++)
        {
          if(isprint(Identif[i][Identif[0].size()-j]) && !isspace(Identif[i][Identif[0].size()-j]))
            flag=1;
          if(isspace(Identif[i][Identif[0].size()-j]) && flag==1)
            Identif[i][Identif[0].size()-j]='_';
        }
      }
      //gleiche Namen
      for(i=0;i<Identif.size();i++)
        for(j=i+1;j<Identif.size();j++)
        {
          flag = 0;
          for(k=0;k<Identif[0].size();k++)
            if(Identif[i][k]!=Identif[j][k])
	      flag = 1;
          if(flag == 0)
          {
            cerr << "Sequences " << i+1 << " and "
	         << j+1 << " have the same name\n";
	    exit(1);
          }
        }
      
}


void dataset::guesstype()
{
     bool Uflag, nonDNAflag;
     unsigned int i, j;
     
     Uflag=nonDNAflag=FALSE;
     for(i=0;i<Seqs.size();i++)
        for(j=0;j<Seqs[i].size();j++) {
	   if(Seqs[i][j]=='U')
	      Uflag=TRUE;
	   else 
	      if(Seqs[i][j]=='E' || Seqs[i][j]=='F' || 
	         Seqs[i][j]=='I' || Seqs[i][j]=='L' ||
	         Seqs[i][j]=='P' || Seqs[i][j]=='Q' ||
	         Seqs[i][j]=='X' || Seqs[i][j]=='Z')
                    nonDNAflag=TRUE;
	}

     if(Uflag && nonDNAflag) {
        cerr << "Don't know what to do: there are U's in\n"
	     << "an amino acid sequence\n";
	exit(EXIT_FAILURE);
     } else {
        if(nonDNAflag)
	  typ=PROTEIN;
	else 
	  typ=DNA;
     }
}


void dataset::print_phylip(ostream& ostr)
{
   unsigned int i, j, Maxspc, Maxseqc;
   
   Maxspc = spez();
   Maxseqc = laen(0);
   
   ostr << setw(10) << Maxspc
        << setw(10) << Maxseqc << endl;
   for(i=0;i<Maxspc;i++)
   {
     for(j=0;j<Maxseqc;j++)
       if(Seqs[i][j]=='_')
          Seqs[i][j]='-';
     for(j=0;j<Identif[i].size() && j<10;j++)
       ostr << Identif[i][j];
     while(j<10) {
       ostr << " ";
       j++;
     }
     ostr << Seqs[i] << endl;	
   }
}

void dataset::print_fasta(ostream& ostr)
{
   unsigned int i, j, charcount;
   
   for(i=0;i<spez();i++)
   {
      ostr<<">";
      for(j=0;j<Identif[i].size();++j)
          ostr<<Identif[i][j];
      ostr<<endl;
      charcount=0;
      for(j=0;j<Seqs[i].size();++j) {
          if(Seqs[i][j]!='-' && Seqs[i][j]!='_') {
	     ++charcount;
             ostr<<Seqs[i][j];
             if(charcount%70==0)
	        ostr<<endl;
	  }
      }
      if(charcount%70!=0)
         ostr<<endl;
   }
}

void dataset::print_screen(ostream& ostr)
{
   vector<string> linewise;
   int i;
   unsigned int j, Maxspc;
   
   Maxspc = spez();
   
   linewise.resize(Maxspc);
   for(j=0;j<Maxspc;j++)
     linewise[j]="";
   i = 0;
   do {
     for(j=0;j<Maxspc;j++)
     {
       linewise[j]+=Seqs[j][i]; 
     }
     i++;
     if(i%70==0){
       for(j=0;j<Maxspc;j++)
       {
          ostr << linewise[j] << endl;
	  linewise[j]=""; 
       }
       ostr << endl;
     }
   } while (i!=laen(0));

   if(linewise[0]!="")
   {
     for(j=0;j<Maxspc;j++)
        ostr << linewise[j] << endl;
     ostr << endl;
   }
}

unsigned int dataset::spez() const
{
  return Identif.size();
}

int dataset::laen(int s) const // gibt Laenge der Sequenz s zurueck (angefangen bei 0)
{
  if((unsigned)s>=Seqs.size())
  {
    cerr << "The specified sequence number is too big.\n"
         << "s=" << s
	 << "There are only " << Seqs.size() << " in the data set.\n"
	 << "Their numbering starts with 0.\n";
    exit(1);
  }
  
  return Seqs[s].size();
}

void dataset::getfreqs(vector<double>& f) // ermittelt absolute Zustandshaeufigkeiten
{
   f.resize(typ);
   for(int i=0;i<typ;i++)
      f[i]=0.0;
   switch(typ) {
      case DNA :
         getACGT(f);
         break;
      case PROTEIN :
         getAAfreqs(f);
         break;
      default :
         cerr << "Unknown data type in function getfreqs\n";
	 exit(EXIT_FAILURE);
   }
}

void dataset::getACGT(vector<double>& f) // ermittelt absolute Basenhaeufigkeiten
{
  unsigned int i, j, k;
  vector<int> states;
  
  for(i=0;i<Identif.size();i++)
    for(j=0;j<Seqs[i].size();j++) {
      chartobase(Seqs[i][j],states);
      for(k=0;k<states.size();k++)
         f[states[k]]+=1.0/(double)(states.size());
    }
}

void dataset::getAAfreqs(vector<double>& f) // ermittelt absolute Aminosaeurehaeufigkeiten
{
  unsigned int i, j, k;
  vector<int> states;
  
  for(i=0;i<Identif.size();i++)
    for(j=0;j<Seqs[i].size();j++) {
      chartoamac(Seqs[i][j],states);
      for(k=0;k<states.size();k++)
         f[states[k]]+=1.0/(double)(states.size());
    }
}


void dataset::push_mult(baum *b, vector<int>& ali_to_node, vector<int>& node_to_ali,
                         vector<vector<char> >& idmat, vector<vector<char> >& seqmat)
{
   vector<vector<vector<int> > > kaus(b->nb.size(),vector<vector<int> >(2)); // Ausrichter der Kinderalignments
   vector<vector<vector<char> > > Namen(b->nb.size());
   vector<vector<vector<char> > > Sequs(b->nb.size());
   unsigned int i, j;
   vector<vector<int> > atn(b->nb.size(),vector<int>(2,-1)),atnalt;
   vector<zustand> zus(b->nb.size());  // enthaelt die Zustaende der paarweisen Alignments
   vector<char> column;  // eine Spalte des zu bastelnden Alignments
   
   if(b==NULL) {
      cerr << "Error in function push_mult: tree points to NULL\n";
      exit(EXIT_FAILURE);
   }
  
   for(i=0;i<b->nb.size();++i) {
     if(!b->nb[i].toroot) {
        push_mult(b->nb[i].nachbar,kaus[i][1],kaus[i][0],Namen[i],Sequs[i]);
     } else {
        kaus[i][0]=b->nb[i].nachbar->nb[b->nb[i].nachbar->whichone(b)].ausrichter;
        kaus[i][1]=b->nb[i].ausrichter;
     }
   }
	
   if(b->isLeaf()) {
     vector<char> vecname;
     for(j=0;j<b->id.name.size();++j)
	 vecname.push_back(b->id.name[j]);
     vector<char> profseq;
     for(j=0;j<b->pro.size();++j)
         profseq.push_back(state_to_char(b->pro[j],typ));
     if(b->isRoot() && b->nb.size()==1) { 
        // idmat und seqmat aus Namen[0], Sequs[0] und eigenem Namen und eigener Sequenz zusammensetzen 
        for(i=0;i<Namen[0].size();++i)
            idmat.push_back(Namen[0][i]);
	idmat.push_back(vecname);
        seqmat.resize(idmat.size());
        do {
	   atnalt=atn;
           zus[0]=iter_zust(kaus[0][1],kaus[0][0],atn[0]);
	   switch(zus[0]) {
	      case START :
	      case ENDE :
	         break;
	      case PAIR :
	         for(i=0;i<Sequs[0].size();++i)
		   seqmat[i].push_back(Sequs[0][i][atnalt[0][1]]);
		 seqmat[i].push_back(profseq[atnalt[0][0]]);
	         break;
	      case DELETION :
	         for(i=0;i<Sequs[0].size();++i)
		   seqmat[i].push_back('_');
		 seqmat[i].push_back(profseq[atnalt[0][0]]);
	         break;
	      case INSERTION :
	         for(i=0;i<Sequs[0].size();++i)
		   seqmat[i].push_back(Sequs[0][i][atnalt[0][1]]);
	         seqmat[i].push_back('_');
	         break;
	      default :
	         cerr << "Unreasonable state in function push_mult\n";
		 exit(EXIT_FAILURE);
	   }
        }while(zus[0]!=ENDE);
     } else {
        // eigenen Namen in idmat und eigene Sequenz in seqmat schreiben
	// ali_to_node ist nb[0].ausrichter
        idmat.push_back(vecname);
        seqmat.push_back(profseq);
	if(b->nb.size()==1) {
           ali_to_node=kaus[0][0];
	   node_to_ali=kaus[0][1];
	} else {  // Entweder hat das Blatt mehrere Vorfahren oder es gibt nur ein Taxon
	   cerr << "Pathological situation\n";
	   exit(EXIT_FAILURE);
	}
     }
   } else {
     for(i=0;i<b->nb.size();++i)
        for(j=0;j<Namen[i].size();++j)
	   idmat.push_back(Namen[i][j]);
     seqmat.resize(idmat.size());
     int aktivins;  // index der im Moment abgearbeiteten Insertion
     int paridx;
     if(!b->isRoot())
        paridx=b->whichone(b->parent());  // Index des parents
     else
        paridx=-2;
     bool nendflag, syncstart;
     
     do {
       atnalt=atn;
       aktivins=-1;
       for(i=0;i<atn.size();++i) 
           zus[i]=iter_zust(kaus[i][1],kaus[i][0],atn[i]);
       nendflag=notatend(zus,syncstart);
       if(nendflag && !syncstart) {  // Wenn man das Ende noch nicht erreicht hat und nicht mehr am Anfang steht
          column.clear();
          for(i=0;i<zus.size() && (aktivins==-1 || aktivins==paridx);++i) // suche erste Insertion
              if(zus[i]==INSERTION)
	         aktivins=(int)i;
	      
          for(i=0;i<zus.size();++i)
              for(j=0;j<Sequs[i].size();++j) { // wird also fuer den parent nicht durchlaufen
	          if(aktivins>-1) {
	            if((int)i!=aktivins)
		       column.push_back('_');
                    else
		       column.push_back(Sequs[i][j][atnalt[i][1]]);
	          } else { // keine Insertionen
	               switch(zus[i]) {
		          case PAIR :
			     column.push_back(Sequs[i][j][atnalt[i][1]]);
			     break;
			  case DELETION :
			     column.push_back('_');
			     break;
                          default :
			     cerr << "Impossible state " <<(int)zus[i]<< " in push_mult\n";
			     exit(EXIT_FAILURE);   
		       }
	          }
          }
	  if(validcolumn(column,idmat.size())) {
	     for(i=0;i<column.size();++i)
	         seqmat[i].push_back(column[i]);
	     if(!b->isRoot()) {
	        if((zus[paridx]==PAIR && aktivins==-1) || (zus[paridx]==INSERTION && aktivins==paridx))
                    //node_to_ali wird weiter unten noch laenger
                    node_to_ali.push_back(ali_to_node.size());
	        else
	            node_to_ali.push_back(ali_to_node.size()-1);
	     }
          }
	  if(!b->isRoot()) {
	     if((zus[paridx]==PAIR && aktivins==-1) || (zus[paridx]==INSERTION && aktivins==paridx))
	       ali_to_node.push_back((int)seqmat[0].size()-1);
          }
          for(i=0;i<atn.size();++i)
             if(aktivins>-1 && (int)i!=aktivins)  // Es gab eine Insertion, die vor den anderen Zustaenden abgearbeitet wurde
	        atn[i]=atnalt[i];
       }
     } while(nendflag);
   }
   
}

bool dataset::validcolumn(vector<char>& col, unsigned int expsize)
{
   if(col.size()!=expsize) {
      cerr << "Alignment column has "<<col.size()<<" entries, yet there should be "<<expsize<<endl;
      exit(EXIT_FAILURE);
   }
   bool retval=TRUE;
   
   for(unsigned int i=0;i<col.size();++i)
       retval=retval && (col[i]=='_');

   return !retval;
}

bool dataset::notatend(vector<zustand>& pzus, bool& syncstart)
{
  bool flag=TRUE;
  syncstart=TRUE;
  
  for(unsigned int i=0;i<pzus.size();++i) {
     flag=flag && (pzus[i]==ENDE);
     syncstart=syncstart && (pzus[i]==START);
     if(pzus[i]==START && syncstart==FALSE || (i>0 && syncstart==TRUE && pzus[i]!=START)) {
        cerr << "Asynchronous start in push_mult\n";
	exit(EXIT_FAILURE);
     }
  }
  
  return !flag;
}


void dataset::row(unsigned int i, vector<char>& rowvec)  // schreibt Spalte i nach rowvec
{
   if((int)i>=laen(0)) {
      cerr << "There is no row "<<i<< " in the alignment\n";
      exit(EXIT_FAILURE);
   }
   rowvec.resize(spez());
   for(unsigned int j=0;j<spez();++j)
      rowvec[j]=Seqs[j][i];
}

ostream& operator<<(ostream& ostr,const vector<char>& d) {
    for (unsigned int i=0; i<d.size(); ++i){
	ostr << d[i];
    }
    return ostr;
}
/*************************************************************/
/****************** classless                      ***********/
/*************************************************************/

zustand get_zust(vector<int>& pAB, vector<int>& pBA, int a, int b)
{
    /* Liefert den zu a und b assoziierten Alignment-Zustand
     */
    if (a==-1) {
	if (b==-1) 
	    return START;
	if (pBA[b]==-1) 
	    return INSERTION;
	return NOSTATE;
    }
    if (a==(int)pAB.size()) {
	if (b==(int)pBA.size())
	    return ENDE;
	return NOSTATE;
    }
    if (b==-1) {
	if (pAB[a]==-1) 
	    return DELETION;
	return NOSTATE;
    }
    if (b==(int)pBA.size()) {
	return NOSTATE;
    }
    if (b==pAB[a]){
	if (a==pBA[b])
	    return PAIR;
	if (pBA[b]<a)
	    return DELETION;
	return NOSTATE;
    }
    if (pBA[b]==a && pAB[a]<b )
	return INSERTION;
    return NOSTATE;

}

zustand iter_zust(vector<int>& AvsB, vector<int>& BvsA, vector<int>& indices)
{
    zustand zust;
    zust=get_zust(AvsB,BvsA,indices[0],indices[1]);
    switch (zust) {
    case START:
    case PAIR:  
    case INSERTION: 
    case DELETION:
	if ( indices[0]==(int)AvsB.size()-1 ) {
	    ++indices[1];
	    if ( indices[1]==(int)BvsA.size() ) 
		++indices[0];
	}
	else {
	    if ( indices[1]==(int)BvsA.size()-1 )
		++indices[0];
	    else {
		if ( AvsB[indices[0]+1]==indices[1] )
		    ++indices[0];
		else {
		    ++indices[1];
		    if ( BvsA[indices[1]]!=indices[0] ) 
			++indices[0];
		}
	    }
	}
	break;
    case ENDE:
	break;
    default:
	cerr << indices[0] << " und " << indices[1] << " ergeben keinen Zustand"
	     << " in iter_zust\n";
	exit(1);
    }
    return zust;
}


