
/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#define DEBUG

#include <string>
#include <vector>
#include <cmath>
#include <cstdlib>
#include <cfloat>  
#include <iostream>

using namespace std;

#include "random.hh"
#include "seqcode.hh"
#include "indel.hh"


#define MAXINTERNALSTRETCH 10
#define CRITICAL_R 0.5
#define TKFBORDER DBL_MIN

/***********************************************************************************************/
/************     ein paar wichtige Funktionen             *************************************/
/***********************************************************************************************/


double untidy_addlog(double a, double b) {  // zweimal im Programm vorhanden!!! (aufraeumen)
    // liefert log(exp(a)+exp(b))
    if (a>b) return a+log(1+exp(b-a));
    else return b+log(1+exp(a-b));
}


indel::indel()
{
  int i;
  //Default constructor; model is TKF1
  model=TKF1;
  params.resize(2);
  params[0]=0.001; //default value of lambda  
  params[1]=0.0011; //default value of mue

  pt = 1.0;
  P.resize(4);
  for(i=0;i<4;i++)
     P[i].resize(5);
  fillP(pt);

  ty = tz = 1.0;
  T.resize(8);
  for(i=0;i<8;i++)
     T[i].resize(9);
  fillT(ty,tz);

  qw = qy = qz = 1.0;
  Q.resize(16);
  for(i=0;i<16;i++)
     Q[i].resize(17);
  fillQ(qw,qy,qz);
  
  huv = hvw = hvx = hxy = hxz = 1.0;
  H.resize(34);
  for(i=0;i<34;i++)
    H[i].resize(35);
  fillH(huv,hvw,hvx,hxy,hxz);
}

indel::indel(double lambda, double mu)
{
  int i;

  //model is TKF1
  model=TKF1;
  params.resize(2);
  params[0]=lambda; //default value of lambda  
  params[1]=mu; //default value of mue  
  if(params[0]>=params[1] || params[0]<0.0) {
     cerr << "In the TKF1 model the insertion rate has to be positive\n"
	  << "and smaller than the deletion rate.\n";
     exit(EXIT_FAILURE);
  }

  pt = 1.0;
  P.resize(4);
  for(i=0;i<4;i++)
     P[i].resize(5);
  fillP(pt);

  ty = tz = 1.0;
  T.resize(8);
  for(i=0;i<8;i++)
     T[i].resize(9);
  fillT(ty,tz);

  qw = qy = qz = 1.0;
  Q.resize(16);
  for(i=0;i<16;i++)
     Q[i].resize(17);
  fillQ(qw,qy,qz);
  
  huv = hvw = hvx = hxy = hxz = 1.0;
  H.resize(34);
  for(i=0;i<34;i++)
    H[i].resize(35);
  fillH(huv,hvw,hvx,hxy,hxz);
}

indel::indel(vector<double> parameter, idarch mod)
{
  int i;

  model = mod;
  switch(model) {
    case TKF1 :
      params.resize(2);

      if(params.size()!=parameter.size()) {
         cerr << "The number of parameters needed by the indel model\n"
              << "and the length of the parameter vector that was passed\n"
	      << "to its constructor are not equal.\n";
         exit(EXIT_FAILURE);
      }
      for(i=0;i<(int)params.size();i++)
        params[i]=parameter[i];
      
      if(params[0]>=params[1] || params[0]<0.0) {
         cerr << "In the TKF1 model the insertion rate has to be positive\n"
	      << "and smaller than the deletion rate.\n";
	 exit(EXIT_FAILURE);
      }
      break;
    case TKF2 :
      params.resize(3);

      if(params.size()!=parameter.size()) {
         cerr << "The number of parameters needed by the indel model\n"
              << "and the length of the parameter vector that was passed\n"
	      << "to its constructor are not equal.\n";
         exit(EXIT_FAILURE);
      }
      for(i=0;i<(int)params.size();i++)
        params[i]=parameter[i];
      
      if(params[0]>=params[1] || params[0]<0.0) {
         cerr << "In the TKF2 model the insertion rate has to be positive\n"
	      << "and smaller than the deletion rate.\n";
	 exit(EXIT_FAILURE);
      }
      
      if(params[2]<=0.0 || params[2]>=1.0) {
         cerr << "The third parameter of the TKF2 model should be positive and larger than 1\n";
	 exit(EXIT_FAILURE);
      }
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
  }
  
  pt = 1.0;
  P.resize(4);
  for(i=0;i<4;i++)
     P[i].resize(5);
  fillP(pt);

  ty = tz = 1.0;
  T.resize(8);
  for(i=0;i<8;i++)
    T[i].resize(9);
  fillT(ty,tz);

  qw = qy = qz = 1.0;
  Q.resize(16);
  for(i=0;i<16;i++)
    Q[i].resize(17);
  fillQ(qw,qy,qz);
  
  huv = hvw = hvx = hxy = hxz = 1.0;
  H.resize(34);
  for(i=0;i<34;i++)
    H[i].resize(35);
  fillH(huv,hvw,hvx,hxy,hxz);
}


idarch indel::returnparams(vector<double> &v)
{
   v.resize(params.size());
   for(unsigned int i=0;i<params.size();i++)
     v[i]=params[i];

   return model;
}

void indel::setparams(vector<double> &newpars)
{
  if(params.size()!=newpars.size()) {
     cerr << "The number of parameters needed by the indel model\n"
          << "and the length of the parameter vector that was passed\n"
	  << "to its constructor are not equal.\n";
     exit(EXIT_FAILURE);
  }
  for(unsigned int i=0;i<params.size();i++)
    params[i]=newpars[i];

  fillP(pt);
  fillT(ty,tz);
  fillQ(qw,qy,qz);
  fillH(huv,hvw,hvx,hxy,hxz);
}


void indel::rescale(double scale)
{
   switch(model) {
      case TKF1:
      case TKF2:
        params[0]*=scale;
	params[1]*=scale;
        break;
      default:
        cerr << "Unknown indel model in function rescale\n";
	exit(EXIT_FAILURE);
   }
}

double indel::tkf1_alpha(double mue, double time)
{
  double expC=exp(log(time)+log(mue));
  double tkfalpha=exp(-expC);
//unsauber:
  if(tkfalpha>1.0-TKFBORDER)
     tkfalpha=1.0-TKFBORDER;
  if(tkfalpha<TKFBORDER)
     tkfalpha=TKFBORDER;
  return tkfalpha;
}

double indel::tkf1_beta(double lambda, double mue, double time)
{
  double expB=exp(log(time)+log(mue-lambda));
  double logA=log(lambda)+log(1.0-exp(-expB))-log(mue-exp(log(lambda)-expB));
  double tkfbeta=exp(logA);
//unsauber:
  if(tkfbeta>1.0-TKFBORDER)
     tkfbeta=1.0-TKFBORDER;
  if(tkfbeta<TKFBORDER)
     tkfbeta=TKFBORDER;
  return tkfbeta;
}

double indel::tkf1_gamma(double lambda, double mue, double time)
{
  double expC=exp(log(time)+log(mue));
  double expB=exp(log(time)+log(mue-lambda));
  double logA=(log(mue)-log(mue-exp(log(lambda)-expB)))+(log(1.0-exp(-expB))-log(1.0-exp(-expC)));
  double tkfgamma=1.0-exp(logA);
//unsauber:
  if(tkfgamma>1.0-TKFBORDER)
     tkfgamma=1.0-TKFBORDER;
  if(tkfgamma<TKFBORDER)
     tkfgamma=TKFBORDER;
  return tkfgamma;
}

double indel::logprob_of_seqlen(unsigned int seqlen)
{
   double l, m, r, lelong, klelong;
   
   switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      lelong=log(l/m);
      klelong=log(1.0-l/m);
      return (double)seqlen*lelong+klelong;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
      if(seqlen==0)
         return log(1.0-l/m);
      else
         return log(1.0-l/m)+log(l/m)+log(1.0-r)+((double)seqlen-1.0)*log((1.0-r)*l/m+r);
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
   }
   
}


void indel::setexplength(unsigned int explen) // veraendert mue so, dass explen die erwartete Sequenzlaenge ist
{
   if(explen==0) {
      cerr << "An expected length of zero does not make sense\n";
      exit(EXIT_FAILURE);
   }
   
   // erwartete Sequenzlaenge bei TKF2=l/((m-l)(1-r))
   double l, m, r;
   
   switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
      break;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
   }
   
   params[1]=l*(1.0+1.0/((double)explen*(1-r)));
}

unsigned int indel::randlen() // zufaellige Sequenzlaenge
{
   // erwartete Sequenzlaenge bei TKF2=l/((m-l)(1-r))
   double l, m, r;
   
   switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
      break;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
   }

   unsigned int retval=0;
   double R = randomunitintervall();
   double logprobsum = log(1.0-l/m), logprob = logprobsum;
   
   while(exp(logprobsum)<R) {
       if(retval==0) {
          logprob += log(l)-log(m)+log(1.0-r);
          logprobsum=untidy_addlog(logprobsum,logprob);
       } else {
          logprob += log(r+(1-r)*l/m);
          logprobsum=untidy_addlog(logprobsum,logprob);
       }
       ++retval;
   }

   return retval;
}

unsigned int indel::explen() // zufaellige Sequenzlaenge
{
   // erwartete Sequenzlaenge bei TKF2=l/((m-l)(1-r))
   double l, m, r;
   
   switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
      break;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
   }

   return (unsigned int)floor(l/((m-l)*(1-r))+0.5);
}


void indel::fillP(double time)
{
  double l, m, r, a, b, g, elong;

  pt = time;
  switch(model) {  // Keine weiteren Modelle mehr einbauen!!!!
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
      break;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
  }
  a=tkf1_alpha(m,time);
  b=tkf1_beta(l,m,time);
  g=tkf1_gamma(l,m,time);
  elong=l/m;
  double la=log(a),
        kla=log(1.0-a),
	 lb=log(b),
	klb=log(1.0-b),
	 lg=log(g),
	klg=log(1.0-g),
     lelong=log(elong),
    klelong=log(1.0-elong),
        klr=log(1.0-r);
    
  P[START][START]=-DBL_MAX;
  P[START][PAIR]=klb+lelong+la;
  P[START][INSERTION]=lb;
  P[START][DELETION]=klb+lelong+kla;
  P[START][ENDE]=klb+klelong;
  P[PAIR][START]=-DBL_MAX;
  P[PAIR][PAIR]=klb+lelong+la;
  P[PAIR][INSERTION]=lb;
  P[PAIR][DELETION]=klb+lelong+kla;
  P[PAIR][ENDE]=klb+klelong;
  P[INSERTION][START]=-DBL_MAX;
  P[INSERTION][PAIR]=klb+lelong+la;
  P[INSERTION][INSERTION]=lb;
  P[INSERTION][DELETION]=klb+lelong+kla;
  P[INSERTION][ENDE]=klb+klelong;
  P[DELETION][START]=-DBL_MAX;
  P[DELETION][PAIR]=klg+lelong+la;
  P[DELETION][INSERTION]=lg;
  P[DELETION][DELETION]=klg+lelong+kla;
  P[DELETION][ENDE]=klg+klelong;
  
  if(model==TKF2) {
     for(int i=0;i<(int)P.size();++i)
       if(i!=START)
         for(int j=0;j<(int)P[i].size();++j) {
	     if(P[i][j]>-DBL_MAX)
	        P[i][j]+=klr;
	     if(i==j)
	        P[i][j]=untidy_addlog(log(r),P[i][j]);
	 } 
  }   
}

void indel::fillT(double txy, double txz)
{
  double l, m, r, a1, b1, g1, a2, b2, g2, elong;

  ty = txy;
  tz = txz;
  switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
      break;
    case TKF2:
      l=params[0];
      m=params[1];
      r=params[2];
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
  }

  a1=tkf1_alpha(m,txy);
  b1=tkf1_beta(l,m,txy);
  g1=tkf1_gamma(l,m,txy);
  a2=tkf1_alpha(m,txz);
  b2=tkf1_beta(l,m,txz);
  g2=tkf1_gamma(l,m,txz);
  elong=l/m;
  double la1=log(a1),
        kla1=log(1.0-a1),
	 lb1=log(b1),
	klb1=log(1.0-b1),
	 lg1=log(g1),
	klg1=log(1.0-g1),
         la2=log(a2),
        kla2=log(1.0-a2),
	 lb2=log(b2),
	klb2=log(1.0-b2),
	 lg2=log(g2),
	klg2=log(1.0-g2),
      lelong=log(elong),
     klelong=log(1.0-elong),
         klr=log(1.0-r);
	      
  T[tSTART][tSTART]=-DBL_MAX;
  T[tSTART][tXYZ]=klb1+klb2+lelong+la2+la1;
  T[tSTART][tXY_]=klb1+klb2+lelong+kla2+la1;
  T[tSTART][tX_Z]=klb1+klb2+lelong+la2+kla1;
  T[tSTART][t__Z]=klb1+lb2;
  T[tSTART][t_Y_1]=lb1;
  T[tSTART][t_Y_2]=-DBL_MAX;
  T[tSTART][tX__]=klb1+klb2+lelong+kla2+kla1;
  T[tSTART][tENDE]=klb1+klb2+klelong;
      
  T[tXYZ][tSTART]=-DBL_MAX;
  T[tXYZ][tXYZ]=klb1+klb2+lelong+la2+la1;
  T[tXYZ][tXY_]=klb1+klb2+lelong+kla2+la1;
  T[tXYZ][tX_Z]=klb1+klb2+lelong+la2+kla1;
  T[tXYZ][t__Z]=klb1+lb2;
  T[tXYZ][t_Y_1]=lb1;
  T[tXYZ][t_Y_2]=-DBL_MAX;
  T[tXYZ][tX__]=klb1+klb2+lelong+kla2+kla1;
  T[tXYZ][tENDE]=klb1+klb2+klelong;
      
  T[tXY_][tSTART]=-DBL_MAX;
  T[tXY_][tXYZ]=klb1+klg2+lelong+la2+la1;
  T[tXY_][tXY_]=klb1+klg2+lelong+kla2+la1;
  T[tXY_][tX_Z]=klb1+klg2+lelong+la2+kla1;
  T[tXY_][t__Z]=klb1+lg2;
  T[tXY_][t_Y_1]=-DBL_MAX;
  T[tXY_][t_Y_2]=lb1;
  T[tXY_][tX__]=klb1+klg2+lelong+kla2+kla1;
  T[tXY_][tENDE]=klb1+klg2+klelong;
      
  T[tX_Z][tSTART]=-DBL_MAX;
  T[tX_Z][tXYZ]=klg1+klb2+lelong+la2+la1;
  T[tX_Z][tXY_]=klg1+klb2+lelong+kla2+la1;
  T[tX_Z][tX_Z]=klg1+klb2+lelong+la2+kla1;
  T[tX_Z][t__Z]=klg1+lb2;
  T[tX_Z][t_Y_1]=lg1;
  T[tX_Z][t_Y_2]=-DBL_MAX;
  T[tX_Z][tX__]=klg1+klb2+lelong+kla2+kla1;
  T[tX_Z][tENDE]=klg1+klb2+klelong;
      
  T[t__Z][tSTART]=-DBL_MAX;
  T[t__Z][tXYZ]=klb2+lelong+la2+la1;
  T[t__Z][tXY_]=klb2+lelong+kla2+la1;
  T[t__Z][tX_Z]=klb2+lelong+la2+kla1;
  T[t__Z][t__Z]=lb2;
  T[t__Z][t_Y_1]=-DBL_MAX;
  T[t__Z][t_Y_2]=-DBL_MAX;
  T[t__Z][tX__]=klb2+lelong+kla2+kla1;
  T[t__Z][tENDE]=klb2+klelong;
      
  T[t_Y_1][tSTART]=-DBL_MAX;
  T[t_Y_1][tXYZ]=klb1+klb2+lelong+la2+la1;
  T[t_Y_1][tXY_]=klb1+klb2+lelong+kla2+la1;
  T[t_Y_1][tX_Z]=klb1+klb2+lelong+la2+kla1;
  T[t_Y_1][t__Z]=klb1+lb2;
  T[t_Y_1][t_Y_1]=lb1;
  T[t_Y_1][t_Y_2]=-DBL_MAX;
  T[t_Y_1][tX__]=klb1+klb2+lelong+kla2+kla1;
  T[t_Y_1][tENDE]=klb1+klb2+klelong;
      
  T[t_Y_2][tSTART]=-DBL_MAX;
  T[t_Y_2][tXYZ]=klb1+klg2+lelong+la2+la1;
  T[t_Y_2][tXY_]=klb1+klg2+lelong+kla2+la1;
  T[t_Y_2][tX_Z]=klb1+klg2+lelong+la2+kla1;
  T[t_Y_2][t__Z]=klb1+lg2;
  T[t_Y_2][t_Y_1]=-DBL_MAX;
  T[t_Y_2][t_Y_2]=lb1;
  T[t_Y_2][tX__]=klb1+klg2+lelong+kla2+kla1;
  T[t_Y_2][tENDE]=klb1+klg2+klelong;
      
  T[tX__][tSTART]=-DBL_MAX;
  T[tX__][tXYZ]=klg1+klg2+lelong+la2+la1;
  T[tX__][tXY_]=klg1+klg2+lelong+kla2+la1;
  T[tX__][tX_Z]=klg1+klg2+lelong+la2+kla1;
  T[tX__][t__Z]=klg1+lg2;
  T[tX__][t_Y_1]=-DBL_MAX;
  T[tX__][t_Y_2]=lg1;
  T[tX__][tX__]=klg1+klg2+lelong+kla2+kla1;
  T[tX__][tENDE]=klg1+klg2+klelong;
      
  if(model==TKF2) {
     for(int i=0;i<(int)T.size();++i)
       if(i!=tSTART)
         for(int j=0;j<(int)T[i].size();++j) {
	     if(T[i][j]>-DBL_MAX)
	        T[i][j]+=klr;
	     if(i==j)
	        T[i][j]=untidy_addlog(log(r),T[i][j]);
	 } 
  }   
}

void indel::fillQ(double twx, double txy, double txz)
{
  double l, m, r, a1, b1, g1, a2, b2, g2, a3, b3, g3, elong;

  qw = twx;
  qy = txy;
  qz = txz;
  switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
      break;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
      break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
  }
  elong=l/m;
  //Indizierung im Holmes-Paper falsch, hier richtig
  a1=tkf1_alpha(m,twx);
  b1=tkf1_beta(l,m,twx);
  g1=tkf1_gamma(l,m,twx);
  a2=tkf1_alpha(m,txz);
  b2=tkf1_beta(l,m,txz);
  g2=tkf1_gamma(l,m,txz);
  a3=tkf1_alpha(m,txy);
  b3=tkf1_beta(l,m,txy);
  g3=tkf1_gamma(l,m,txy);
  double la1=log(a1),
        kla1=log(1.0-a1),
	 lb1=log(b1),
	klb1=log(1.0-b1),
	 lg1=log(g1),
	klg1=log(1.0-g1),
         la2=log(a2),
        kla2=log(1.0-a2),
	 lb2=log(b2),
	klb2=log(1.0-b2),
	 lg2=log(g2),
	klg2=log(1.0-g2),
         la3=log(a3),
        kla3=log(1.0-a3),
	 lb3=log(b3),
	klb3=log(1.0-b3),
	 lg3=log(g3),
	klg3=log(1.0-g3),
      lelong=log(elong),
     klelong=log(1.0-elong),
         klr=log(1.0-r);
	      
  Q[qSTART][qWXYZ]=klb3+klb2+klb1+lelong+la1+la2+la3;
  Q[qSTART][qWXY_]=klb3+klb2+klb1+lelong+la1+kla2+la3;
  Q[qSTART][qWX_Z]=klb3+klb2+klb1+lelong+la1+la2+kla3;
  Q[qSTART][q_XYZ]=klb3+klb2+lb1+la2+la3;
  Q[qSTART][qWX__]=klb3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[qSTART][q_XY_]=klb3+klb2+lb1+kla2+la3;
  Q[qSTART][q_X_Z]=klb3+klb2+lb1+la2+kla3;
  Q[qSTART][qW___]=klb3+klb2+klb1+lelong+kla1;
  Q[qSTART][q_X__]=klb3+klb2+lb1+kla2+kla3;
  Q[qSTART][q__Y_1]=lb3;
  Q[qSTART][q__Y_2]=-DBL_MAX;
  Q[qSTART][q__Y_3]=-DBL_MAX;
  Q[qSTART][q__Y_4]=-DBL_MAX;
  Q[qSTART][q___Z1]=klb3+lb2;
  Q[qSTART][q___Z2]=-DBL_MAX;
  Q[qSTART][qSTART]=-DBL_MAX;
  Q[qSTART][qENDE]=klb3+klb2+klb1+klelong;
      
  Q[qWXYZ][qWXYZ]=klb3+klb2+klb1+lelong+la1+la2+la3;
  Q[qWXYZ][qWXY_]=klb3+klb2+klb1+lelong+la1+kla2+la3;
  Q[qWXYZ][qWX_Z]=klb3+klb2+klb1+lelong+la1+la2+kla3;
  Q[qWXYZ][q_XYZ]=klb3+klb2+lb1+la2+la3;
  Q[qWXYZ][qWX__]=klb3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[qWXYZ][q_XY_]=klb3+klb2+lb1+kla2+la3;
  Q[qWXYZ][q_X_Z]=klb3+klb2+lb1+la2+kla3;
  Q[qWXYZ][qW___]=klb3+klb2+klb1+lelong+kla1;
  Q[qWXYZ][q_X__]=klb3+klb2+lb1+kla2+kla3;
  Q[qWXYZ][q__Y_1]=-DBL_MAX;
  Q[qWXYZ][q__Y_2]=-DBL_MAX;
  Q[qWXYZ][q__Y_3]=lb3;
  Q[qWXYZ][q__Y_4]=-DBL_MAX;
  Q[qWXYZ][q___Z1]=-DBL_MAX;
  Q[qWXYZ][q___Z2]=klb3+lb2;
  Q[qWXYZ][qSTART]=-DBL_MAX;
  Q[qWXYZ][qENDE]=klb3+klb2+klb1+klelong;

  Q[qWXY_][qWXYZ]=klb3+klg2+klb1+lelong+la1+la2+la3;
  Q[qWXY_][qWXY_]=klb3+klg2+klb1+lelong+la1+kla2+la3;
  Q[qWXY_][qWX_Z]=klb3+klg2+klb1+lelong+la1+la2+kla3;
  Q[qWXY_][q_XYZ]=klb3+klg2+lb1+la2+la3;
  Q[qWXY_][qWX__]=klb3+klg2+klb1+lelong+la1+kla2+kla3;
  Q[qWXY_][q_XY_]=klb3+klg2+lb1+kla2+la3;
  Q[qWXY_][q_X_Z]=klb3+klg2+lb1+la2+kla3;
  Q[qWXY_][qW___]=klb3+klg2+klb1+lelong+kla1;
  Q[qWXY_][q_X__]=klb3+klg2+lb1+kla2+kla3;
  Q[qWXY_][q__Y_1]=-DBL_MAX;
  Q[qWXY_][q__Y_2]=-DBL_MAX;
  Q[qWXY_][q__Y_3]=-DBL_MAX;
  Q[qWXY_][q__Y_4]=lb3;
  Q[qWXY_][q___Z1]=-DBL_MAX;
  Q[qWXY_][q___Z2]=klb3+lg2;
  Q[qWXY_][qSTART]=-DBL_MAX;
  Q[qWXY_][qENDE]=klb3+klg2+klb1+klelong;

  Q[qWX_Z][qWXYZ]=klg3+klb2+klb1+lelong+la1+la2+la3;
  Q[qWX_Z][qWXY_]=klg3+klb2+klb1+lelong+la1+kla2+la3;
  Q[qWX_Z][qWX_Z]=klg3+klb2+klb1+lelong+la1+la2+kla3;
  Q[qWX_Z][q_XYZ]=klg3+klb2+lb1+la2+la3;
  Q[qWX_Z][qWX__]=klg3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[qWX_Z][q_XY_]=klg3+klb2+lb1+kla2+la3;
  Q[qWX_Z][q_X_Z]=klg3+klb2+lb1+la2+kla3;
  Q[qWX_Z][qW___]=klg3+klb2+klb1+lelong+kla1;
  Q[qWX_Z][q_X__]=klg3+klb2+lb1+kla2+kla3;
  Q[qWX_Z][q__Y_1]=-DBL_MAX;
  Q[qWX_Z][q__Y_2]=-DBL_MAX;
  Q[qWX_Z][q__Y_3]=lg3;
  Q[qWX_Z][q__Y_4]=-DBL_MAX;
  Q[qWX_Z][q___Z1]=-DBL_MAX;
  Q[qWX_Z][q___Z2]=klg3+lb2;
  Q[qWX_Z][qSTART]=-DBL_MAX;
  Q[qWX_Z][qENDE]=klg3+klb2+klb1+klelong;

  Q[q_XYZ][qWXYZ]=klb3+klb2+klb1+lelong+la1+la2+la3;
  Q[q_XYZ][qWXY_]=klb3+klb2+klb1+lelong+la1+kla2+la3;
  Q[q_XYZ][qWX_Z]=klb3+klb2+klb1+lelong+la1+la2+kla3;
  Q[q_XYZ][q_XYZ]=klb3+klb2+lb1+la2+la3;
  Q[q_XYZ][qWX__]=klb3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[q_XYZ][q_XY_]=klb3+klb2+lb1+kla2+la3;
  Q[q_XYZ][q_X_Z]=klb3+klb2+lb1+la2+kla3;
  Q[q_XYZ][qW___]=klb3+klb2+klb1+lelong+kla1;
  Q[q_XYZ][q_X__]=klb3+klb2+lb1+kla2+kla3;
  Q[q_XYZ][q__Y_1]=lb3;
  Q[q_XYZ][q__Y_2]=-DBL_MAX;
  Q[q_XYZ][q__Y_3]=-DBL_MAX;
  Q[q_XYZ][q__Y_4]=-DBL_MAX;
  Q[q_XYZ][q___Z1]=klb3+lb2;
  Q[q_XYZ][q___Z2]=-DBL_MAX;
  Q[q_XYZ][qSTART]=-DBL_MAX;
  Q[q_XYZ][qENDE]=klb3+klb2+klb1+klelong;

  Q[qWX__][qWXYZ]=klg3+klg2+klb1+lelong+la1+la2+la3;
  Q[qWX__][qWXY_]=klg3+klg2+klb1+lelong+la1+kla2+la3;
  Q[qWX__][qWX_Z]=klg3+klg2+klb1+lelong+la1+la2+kla3;
  Q[qWX__][q_XYZ]=klg3+klg2+lb1+la2+la3;
  Q[qWX__][qWX__]=klg3+klg2+klb1+lelong+la1+kla2+kla3;
  Q[qWX__][q_XY_]=klg3+klg2+lb1+kla2+la3;
  Q[qWX__][q_X_Z]=klg3+klg2+lb1+la2+kla3;
  Q[qWX__][qW___]=klg3+klg2+klb1+lelong+kla1;
  Q[qWX__][q_X__]=klg3+klg2+lb1+kla2+kla3;
  Q[qWX__][q__Y_1]=-DBL_MAX;
  Q[qWX__][q__Y_2]=-DBL_MAX;
  Q[qWX__][q__Y_3]=-DBL_MAX;
  Q[qWX__][q__Y_4]=lg3;
  Q[qWX__][q___Z1]=-DBL_MAX;
  Q[qWX__][q___Z2]=klg3+lg2;
  Q[qWX__][qSTART]=-DBL_MAX;
  Q[qWX__][qENDE]=klg3+klg2+klb1+klelong;

  Q[q_XY_][qWXYZ]=klb3+klg2+klb1+lelong+la1+la2+la3;
  Q[q_XY_][qWXY_]=klb3+klg2+klb1+lelong+la1+kla2+la3;
  Q[q_XY_][qWX_Z]=klb3+klg2+klb1+lelong+la1+la2+kla3;
  Q[q_XY_][q_XYZ]=klb3+klg2+lb1+la2+la3;
  Q[q_XY_][qWX__]=klb3+klg2+klb1+lelong+la1+kla2+kla3;
  Q[q_XY_][q_XY_]=klb3+klg2+lb1+kla2+la3;
  Q[q_XY_][q_X_Z]=klb3+klg2+lb1+la2+kla3;
  Q[q_XY_][qW___]=klb3+klg2+klb1+lelong+kla1;
  Q[q_XY_][q_X__]=klb3+klg2+lb1+kla2+kla3;
  Q[q_XY_][q__Y_1]=-DBL_MAX;
  Q[q_XY_][q__Y_2]=lb3;
  Q[q_XY_][q__Y_3]=-DBL_MAX;
  Q[q_XY_][q__Y_4]=-DBL_MAX;
  Q[q_XY_][q___Z1]=klb3+lg2;
  Q[q_XY_][q___Z2]=-DBL_MAX;
  Q[q_XY_][qSTART]=-DBL_MAX;
  Q[q_XY_][qENDE]=klb3+klg2+klb1+klelong;

  Q[q_X_Z][qWXYZ]=klg3+klb2+klb1+lelong+la1+la2+la3;
  Q[q_X_Z][qWXY_]=klg3+klb2+klb1+lelong+la1+kla2+la3;
  Q[q_X_Z][qWX_Z]=klg3+klb2+klb1+lelong+la1+la2+kla3;
  Q[q_X_Z][q_XYZ]=klg3+klb2+lb1+la2+la3;
  Q[q_X_Z][qWX__]=klg3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[q_X_Z][q_XY_]=klg3+klb2+lb1+kla2+la3;
  Q[q_X_Z][q_X_Z]=klg3+klb2+lb1+la2+kla3;
  Q[q_X_Z][qW___]=klg3+klb2+klb1+lelong+kla1;
  Q[q_X_Z][q_X__]=klg3+klb2+lb1+kla2+kla3;
  Q[q_X_Z][q__Y_1]=lg3;
  Q[q_X_Z][q__Y_2]=-DBL_MAX;
  Q[q_X_Z][q__Y_3]=-DBL_MAX;
  Q[q_X_Z][q__Y_4]=-DBL_MAX;
  Q[q_X_Z][q___Z1]=klg3+lb2;
  Q[q_X_Z][q___Z2]=-DBL_MAX;
  Q[q_X_Z][qSTART]=-DBL_MAX;
  Q[q_X_Z][qENDE]=klg3+klb2+klb1+klelong;

  Q[qW___][qWXYZ]=klg1+lelong+la1+la2+la3;
  Q[qW___][qWXY_]=klg1+lelong+la1+kla2+la3;
  Q[qW___][qWX_Z]=klg1+lelong+la1+la2+kla3;
  Q[qW___][q_XYZ]=lg1+la2+la3;
  Q[qW___][qWX__]=klg1+lelong+la1+kla2+kla3;
  Q[qW___][q_XY_]=lg1+kla2+la3;
  Q[qW___][q_X_Z]=lg1+la2+kla3;
  Q[qW___][qW___]=klg1+lelong+kla1;
  Q[qW___][q_X__]=lg1+kla2+kla3;
  Q[qW___][q__Y_1]=-DBL_MAX;
  Q[qW___][q__Y_2]=-DBL_MAX;
  Q[qW___][q__Y_3]=-DBL_MAX;
  Q[qW___][q__Y_4]=-DBL_MAX;
  Q[qW___][q___Z1]=-DBL_MAX;
  Q[qW___][q___Z2]=-DBL_MAX;
  Q[qW___][qSTART]=-DBL_MAX;
  Q[qW___][qENDE]=klg1+klelong;

  Q[q_X__][qWXYZ]=klg3+klg2+klb1+lelong+la1+la2+la3;
  Q[q_X__][qWXY_]=klg3+klg2+klb1+lelong+la1+kla2+la3;
  Q[q_X__][qWX_Z]=klg3+klg2+klb1+lelong+la1+la2+kla3;
  Q[q_X__][q_XYZ]=klg3+klg2+lb1+la2+la3;
  Q[q_X__][qWX__]=klg3+klg2+klb1+lelong+la1+kla2+kla3;
  Q[q_X__][q_XY_]=klg3+klg2+lb1+kla2+la3;
  Q[q_X__][q_X_Z]=klg3+klg2+lb1+la2+kla3;
  Q[q_X__][qW___]=klg3+klg2+klb1+lelong+kla1;
  Q[q_X__][q_X__]=klg3+klg2+lb1+kla2+kla3;
  Q[q_X__][q__Y_1]=-DBL_MAX;
  Q[q_X__][q__Y_2]=lg3;
  Q[q_X__][q__Y_3]=-DBL_MAX;
  Q[q_X__][q__Y_4]=-DBL_MAX;
  Q[q_X__][q___Z1]=klg3+lg2;
  Q[q_X__][q___Z2]=-DBL_MAX;
  Q[q_X__][qSTART]=-DBL_MAX;
  Q[q_X__][qENDE]=klg3+klg2+klb1+klelong;

  Q[q__Y_1][qWXYZ]=klb3+klb2+klb1+lelong+la1+la2+la3;
  Q[q__Y_1][qWXY_]=klb3+klb2+klb1+lelong+la1+kla2+la3;
  Q[q__Y_1][qWX_Z]=klb3+klb2+klb1+lelong+la1+la2+kla3;
  Q[q__Y_1][q_XYZ]=klb3+klb2+lb1+la2+la3;
  Q[q__Y_1][qWX__]=klb3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[q__Y_1][q_XY_]=klb3+klb2+lb1+kla2+la3;
  Q[q__Y_1][q_X_Z]=klb3+klb2+lb1+la2+kla3;
  Q[q__Y_1][qW___]=klb3+klb2+klb1+lelong+kla1;
  Q[q__Y_1][q_X__]=klb3+klb2+lb1+kla2+kla3;
  Q[q__Y_1][q__Y_1]=lb3;
  Q[q__Y_1][q__Y_2]=-DBL_MAX;
  Q[q__Y_1][q__Y_3]=-DBL_MAX;
  Q[q__Y_1][q__Y_4]=-DBL_MAX;
  Q[q__Y_1][q___Z1]=klb3+lb2;
  Q[q__Y_1][q___Z2]=-DBL_MAX;
  Q[q__Y_1][qSTART]=-DBL_MAX;
  Q[q__Y_1][qENDE]=klb3+klb2+klb1+klelong;

  Q[q__Y_2][qWXYZ]=klb3+klg2+klb1+lelong+la1+la2+la3;
  Q[q__Y_2][qWXY_]=klb3+klg2+klb1+lelong+la1+kla2+la3;
  Q[q__Y_2][qWX_Z]=klb3+klg2+klb1+lelong+la1+la2+kla3;
  Q[q__Y_2][q_XYZ]=klb3+klg2+lb1+la2+la3;
  Q[q__Y_2][qWX__]=klb3+klg2+klb1+lelong+la1+kla2+kla3;
  Q[q__Y_2][q_XY_]=klb3+klg2+lb1+kla2+la3;
  Q[q__Y_2][q_X_Z]=klb3+klg2+lb1+la2+kla3;
  Q[q__Y_2][qW___]=klb3+klg2+klb1+lelong+kla1;
  Q[q__Y_2][q_X__]=klb3+klg2+lb1+kla2+kla3;
  Q[q__Y_2][q__Y_1]=-DBL_MAX;
  Q[q__Y_2][q__Y_2]=lb3;
  Q[q__Y_2][q__Y_3]=-DBL_MAX;
  Q[q__Y_2][q__Y_4]=-DBL_MAX;
  Q[q__Y_2][q___Z1]=klb3+lg2;
  Q[q__Y_2][q___Z2]=-DBL_MAX;
  Q[q__Y_2][qSTART]=-DBL_MAX;
  Q[q__Y_2][qENDE]=klb3+klg2+klb1+klelong;

  Q[q__Y_3][qWXYZ]=klb3+klb2+klb1+lelong+la1+la2+la3;
  Q[q__Y_3][qWXY_]=klb3+klb2+klb1+lelong+la1+kla2+la3;
  Q[q__Y_3][qWX_Z]=klb3+klb2+klb1+lelong+la1+la2+kla3;
  Q[q__Y_3][q_XYZ]=klb3+klb2+lb1+la2+la3;
  Q[q__Y_3][qWX__]=klb3+klb2+klb1+lelong+la1+kla2+kla3;
  Q[q__Y_3][q_XY_]=klb3+klb2+lb1+kla2+la3;
  Q[q__Y_3][q_X_Z]=klb3+klb2+lb1+la2+kla3;
  Q[q__Y_3][qW___]=klb3+klb2+klb1+lelong+kla1;
  Q[q__Y_3][q_X__]=klb3+klb2+lb1+kla2+kla3;
  Q[q__Y_3][q__Y_1]=-DBL_MAX;
  Q[q__Y_3][q__Y_2]=-DBL_MAX;
  Q[q__Y_3][q__Y_3]=lb3;
  Q[q__Y_3][q__Y_4]=-DBL_MAX;
  Q[q__Y_3][q___Z1]=-DBL_MAX;
  Q[q__Y_3][q___Z2]=klb3+lb2;
  Q[q__Y_3][qSTART]=-DBL_MAX;
  Q[q__Y_3][qENDE]=klb3+klb2+klb1+klelong;

  Q[q__Y_4][qWXYZ]=klb3+klg2+klb1+lelong+la1+la2+la3;
  Q[q__Y_4][qWXY_]=klb3+klg2+klb1+lelong+la1+kla2+la3;
  Q[q__Y_4][qWX_Z]=klb3+klg2+klb1+lelong+la1+la2+kla3;
  Q[q__Y_4][q_XYZ]=klb3+klg2+lb1+la2+la3;
  Q[q__Y_4][qWX__]=klb3+klg2+klb1+lelong+la1+kla2+kla3;
  Q[q__Y_4][q_XY_]=klb3+klg2+lb1+kla2+la3;
  Q[q__Y_4][q_X_Z]=klb3+klg2+lb1+la2+kla3;
  Q[q__Y_4][qW___]=klb3+klg2+klb1+lelong+kla1;
  Q[q__Y_4][q_X__]=klb3+klg2+lb1+kla2+kla3;
  Q[q__Y_4][q__Y_1]=-DBL_MAX;
  Q[q__Y_4][q__Y_2]=-DBL_MAX;
  Q[q__Y_4][q__Y_3]=-DBL_MAX;
  Q[q__Y_4][q__Y_4]=lb3;
  Q[q__Y_4][q___Z1]=-DBL_MAX;
  Q[q__Y_4][q___Z2]=klb3+lg2;
  Q[q__Y_4][qSTART]=-DBL_MAX;
  Q[q__Y_4][qENDE]=klb3+klg2+klb1+klelong;

  Q[q___Z1][qWXYZ]=klb2+klb1+lelong+la1+la2+la3;
  Q[q___Z1][qWXY_]=klb2+klb1+lelong+la1+kla2+la3;
  Q[q___Z1][qWX_Z]=klb2+klb1+lelong+la1+la2+kla3;
  Q[q___Z1][q_XYZ]=klb2+lb1+la2+la3;
  Q[q___Z1][qWX__]=klb2+klb1+lelong+la1+kla2+kla3;
  Q[q___Z1][q_XY_]=klb2+lb1+kla2+la3;
  Q[q___Z1][q_X_Z]=klb2+lb1+la2+kla3;
  Q[q___Z1][qW___]=klb2+klb1+lelong+kla1;
  Q[q___Z1][q_X__]=klb2+lb1+kla2+kla3;
  Q[q___Z1][q__Y_1]=-DBL_MAX;
  Q[q___Z1][q__Y_2]=-DBL_MAX;
  Q[q___Z1][q__Y_3]=-DBL_MAX;
  Q[q___Z1][q__Y_4]=-DBL_MAX;
  Q[q___Z1][q___Z1]=lb2;
  Q[q___Z1][q___Z2]=-DBL_MAX;
  Q[q___Z1][qSTART]=-DBL_MAX;
  Q[q___Z1][qENDE]=klb2+klb1+klelong;

  Q[q___Z2][qWXYZ]=klb2+klb1+lelong+la1+la2+la3;
  Q[q___Z2][qWXY_]=klb2+klb1+lelong+la1+kla2+la3;
  Q[q___Z2][qWX_Z]=klb2+klb1+lelong+la1+la2+kla3;
  Q[q___Z2][q_XYZ]=klb2+lb1+la2+la3;
  Q[q___Z2][qWX__]=klb2+klb1+lelong+la1+kla2+kla3;
  Q[q___Z2][q_XY_]=klb2+lb1+kla2+la3;
  Q[q___Z2][q_X_Z]=klb2+lb1+la2+kla3;
  Q[q___Z2][qW___]=klb2+klb1+lelong+kla1;
  Q[q___Z2][q_X__]=klb2+lb1+kla2+kla3;
  Q[q___Z2][q__Y_1]=-DBL_MAX;
  Q[q___Z2][q__Y_2]=-DBL_MAX;
  Q[q___Z2][q__Y_3]=-DBL_MAX;
  Q[q___Z2][q__Y_4]=-DBL_MAX;
  Q[q___Z2][q___Z1]=-DBL_MAX;
  Q[q___Z2][q___Z2]=lb2;
  Q[q___Z2][qSTART]=-DBL_MAX;
  Q[q___Z2][qENDE]=klb2+klb1+klelong;

  if(model==TKF2) {
     for(int i=0;i<(int)Q.size();++i)
       if(i!=qSTART)
         for(int j=0;j<(int)Q[i].size();++j) {
	     if(Q[i][j]>-DBL_MAX)
	        Q[i][j]+=klr;
	     if(i==j)
	        Q[i][j]=untidy_addlog(log(r),Q[i][j]);
	 } 
  }   
}

void indel::fillH(double tuv, double tvw, double tvx, double txy, double txz)
{
  double l, m, r, a1, b1, g1, a2, b2, g2, a3, b3, g3, a4, b4, g4, a5, b5, g5, elong;
  unsigned int i, j;
  huv = tuv;
  hvw = tvw;
  hvx = tvx;
  hxy = txy;
  hxz = txz;
  switch(model) {
    case TKF1 :
      l=params[0];
      m=params[1];
      r=0.0;
     break;
    case TKF2 :
      l=params[0];
      m=params[1];
      r=params[2];
     break;
    default :
      cerr << "Unknown indel model\n";
      exit(EXIT_FAILURE);
  }
 
  for(i=0;i<H.size();i++)
     for(j=0;j<H[i].size();j++)
        H[i][j]=-DBL_MAX;

  elong=l/m;
  a1=tkf1_alpha(m,huv);
  b1=tkf1_beta(l,m,huv);
  g1=tkf1_gamma(l,m,huv);
  a2=tkf1_alpha(m,hvw);
  b2=tkf1_beta(l,m,hvw);
  g2=tkf1_gamma(l,m,hvw);
  a3=tkf1_alpha(m,hvx);
  b3=tkf1_beta(l,m,hvx);
  g3=tkf1_gamma(l,m,hvx);
  a4=tkf1_alpha(m,hxy);
  b4=tkf1_beta(l,m,hxy);
  g4=tkf1_gamma(l,m,hxy);
  a5=tkf1_alpha(m,hxz);
  b5=tkf1_beta(l,m,hxz);
  g5=tkf1_gamma(l,m,hxz);
  double la1=log(a1),
        kla1=log(1.0-a1),
	 lb1=log(b1),
	klb1=log(1.0-b1),
	 lg1=log(g1),
	klg1=log(1.0-g1),
         la2=log(a2),
        kla2=log(1.0-a2),
	 lb2=log(b2),
	klb2=log(1.0-b2),
	 lg2=log(g2),
	klg2=log(1.0-g2),
         la3=log(a3),
        kla3=log(1.0-a3),
	 lb3=log(b3),
	klb3=log(1.0-b3),
	 lg3=log(g3),
	klg3=log(1.0-g3),
         la4=log(a4),
        kla4=log(1.0-a4),
	 lb4=log(b4),
	klb4=log(1.0-b4),
	 lg4=log(g4),
	klg4=log(1.0-g4),
         la5=log(a5),
        kla5=log(1.0-a5),
	 lb5=log(b5),
	klb5=log(1.0-b5),
	 lg5=log(g5),
	klg5=log(1.0-g5),
      lelong=log(elong),
     klelong=log(1.0-elong),
         klr=log(1.0-r);
	      
  H[hSTART][hW1] = lb2;
  H[hSTART][hY1] = klb2+lb4;	
  H[hSTART][h_____Z] = klb2+klb4+lb5;	
  H[hSTART][h___XYZ] = klb2+klb4+klb5+lb3+la5+la4;	
  H[hSTART][h___X_Z] = klb2+klb4+klb5+lb3+la5+kla4;	
  H[hSTART][h___XY_] = klb2+klb4+klb5+lb3+kla5+la4;	
  H[hSTART][h___X__] = klb2+klb4+klb5+lb3+kla5+kla4;	
  H[hSTART][hUVWXYZ] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hSTART][hUV_XYZ] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hSTART][hUVWX_Z] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hSTART][hUV_X_Z] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hSTART][hUVWXY_] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hSTART][hUV_XY_] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hSTART][hUVWX__] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hSTART][hUV_X__] = klb2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hSTART][hUVW___] = klb2+klb4+klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hSTART][hUV____] = klb2+klb4+klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hSTART][h_VWXYZ] = klb2+klb4+klb5+klb3+lb1+la3+la5+la4+la2;	
  H[hSTART][h_V_XYZ] = klb2+klb4+klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[hSTART][h_VWX_Z] = klb2+klb4+klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[hSTART][h_V_X_Z] = klb2+klb4+klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hSTART][h_VWXY_] = klb2+klb4+klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[hSTART][h_V_XY_] = klb2+klb4+klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hSTART][h_VWX__] = klb2+klb4+klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hSTART][h_V_X__] = klb2+klb4+klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hSTART][h_VW___] = klb2+klb4+klb5+klb3+lb1+kla3+la2;	
  H[hSTART][h_V____] = klb2+klb4+klb5+klb3+lb1+kla3+kla2;	
  H[hSTART][hU_____] = klb2+klb4+klb5+klb3+klb1+lelong+kla1;	
  H[hSTART][hENDE]   = klb2+klb4+klb5+klb3+klb1+klelong;
  for(j=0;j<H[hSTART].size();j++) {
    H[hUVWXYZ][j]=H[hSTART][j]; 	
    H[h_VWXYZ][j]=H[hSTART][j]; 
    H[hW1][j]=H[hSTART][j]; 
  } 	
        

  H[h___XYZ][hY1] = lb4;	
  H[h___XYZ][h_____Z] = klb4+lb5;	
  H[h___XYZ][h___XYZ] = klb4+klb5+lb3+la5+la4;	
  H[h___XYZ][h___X_Z] = klb4+klb5+lb3+la5+kla4;	
  H[h___XYZ][h___XY_] = klb4+klb5+lb3+kla5+la4;	
  H[h___XYZ][h___X__] = klb4+klb5+lb3+kla5+kla4;	
  H[h___XYZ][hUVWXYZ] = klb4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[h___XYZ][hUV_XYZ] = klb4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[h___XYZ][hUVWX_Z] = klb4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[h___XYZ][hUV_X_Z] = klb4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[h___XYZ][hUVWXY_] = klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[h___XYZ][hUV_XY_] = klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[h___XYZ][hUVWX__] = klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[h___XYZ][hUV_X__] = klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[h___XYZ][hUVW___] = klb4+klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[h___XYZ][hUV____] = klb4+klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[h___XYZ][h_VWXYZ] = klb4+klb5+klb3+lb1+la3+la5+la4+la2;	
  H[h___XYZ][h_V_XYZ] = klb4+klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[h___XYZ][h_VWX_Z] = klb4+klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[h___XYZ][h_V_X_Z] = klb4+klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[h___XYZ][h_VWXY_] = klb4+klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[h___XYZ][h_V_XY_] = klb4+klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[h___XYZ][h_VWX__] = klb4+klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[h___XYZ][h_V_X__] = klb4+klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[h___XYZ][h_VW___] = klb4+klb5+klb3+lb1+kla3+la2;	
  H[h___XYZ][h_V____] = klb4+klb5+klb3+lb1+kla3+kla2;	
  H[h___XYZ][hU_____] = klb4+klb5+klb3+klb1+lelong+kla1;	
  H[h___XYZ][hENDE]   = klb4+klb5+klb3+klb1+klelong;
  for(j=0;j<H[h___XYZ].size();j++)
    H[hY1][j]=H[h___XYZ][j]; 	
        

  H[hUV_XYZ][hW1] = lg2;
  H[hUV_XYZ][hY1] = klg2+lb4;	
  H[hUV_XYZ][h_____Z] = klg2+klb4+lb5;	
  H[hUV_XYZ][h___XYZ] = klg2+klb4+klb5+lb3+la5+la4;	
  H[hUV_XYZ][h___X_Z] = klg2+klb4+klb5+lb3+la5+kla4;	
  H[hUV_XYZ][h___XY_] = klg2+klb4+klb5+lb3+kla5+la4;	
  H[hUV_XYZ][h___X__] = klg2+klb4+klb5+lb3+kla5+kla4;	
  H[hUV_XYZ][hUVWXYZ] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUV_XYZ][hUV_XYZ] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUV_XYZ][hUVWX_Z] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUV_XYZ][hUV_X_Z] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUV_XYZ][hUVWXY_] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUV_XYZ][hUV_XY_] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUV_XYZ][hUVWX__] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUV_XYZ][hUV_X__] = klg2+klb4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUV_XYZ][hUVW___] = klg2+klb4+klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUV_XYZ][hUV____] = klg2+klb4+klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUV_XYZ][h_VWXYZ] = klg2+klb4+klb5+klb3+lb1+la3+la5+la4+la2;	
  H[hUV_XYZ][h_V_XYZ] = klg2+klb4+klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUV_XYZ][h_VWX_Z] = klg2+klb4+klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUV_XYZ][h_V_X_Z] = klg2+klb4+klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUV_XYZ][h_VWXY_] = klg2+klb4+klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUV_XYZ][h_V_XY_] = klg2+klb4+klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUV_XYZ][h_VWX__] = klg2+klb4+klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUV_XYZ][h_V_X__] = klg2+klb4+klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUV_XYZ][h_VW___] = klg2+klb4+klb5+klb3+lb1+kla3+la2;	
  H[hUV_XYZ][h_V____] = klg2+klb4+klb5+klb3+lb1+kla3+kla2;	
  H[hUV_XYZ][hU_____] = klg2+klb4+klb5+klb3+klb1+lelong+kla1;	
  H[hUV_XYZ][hENDE]   = klg2+klb4+klb5+klb3+klb1+klelong;
  for(j=0;j<H[hUV_XYZ].size();j++)
     H[h_V_XYZ][j]=H[hUV_XYZ][j];
     
  H[hUVWX_Z][hW2] = lb2;
  H[hUVWX_Z][hY1] = klb2+lg4;	
  H[hUVWX_Z][h_____Z] = klb2+klg4+lb5;	
  H[hUVWX_Z][h___XYZ] = klb2+klg4+klb5+lb3+la5+la4;	
  H[hUVWX_Z][h___X_Z] = klb2+klg4+klb5+lb3+la5+kla4;	
  H[hUVWX_Z][h___XY_] = klb2+klg4+klb5+lb3+kla5+la4;	
  H[hUVWX_Z][h___X__] = klb2+klg4+klb5+lb3+kla5+kla4;	
  H[hUVWX_Z][hUVWXYZ] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUVWX_Z][hUV_XYZ] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUVWX_Z][hUVWX_Z] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUVWX_Z][hUV_X_Z] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUVWX_Z][hUVWXY_] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUVWX_Z][hUV_XY_] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUVWX_Z][hUVWX__] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUVWX_Z][hUV_X__] = klb2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUVWX_Z][hUVW___] = klb2+klg4+klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUVWX_Z][hUV____] = klb2+klg4+klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUVWX_Z][h_VWXYZ] = klb2+klg4+klb5+klb3+lb1+la3+la5+la4+la2;	
  H[hUVWX_Z][h_V_XYZ] = klb2+klg4+klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUVWX_Z][h_VWX_Z] = klb2+klg4+klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUVWX_Z][h_V_X_Z] = klb2+klg4+klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUVWX_Z][h_VWXY_] = klb2+klg4+klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUVWX_Z][h_V_XY_] = klb2+klg4+klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUVWX_Z][h_VWX__] = klb2+klg4+klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUVWX_Z][h_V_X__] = klb2+klg4+klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUVWX_Z][h_VW___] = klb2+klg4+klb5+klb3+lb1+kla3+la2;	
  H[hUVWX_Z][h_V____] = klb2+klg4+klb5+klb3+lb1+kla3+kla2;	
  H[hUVWX_Z][hU_____] = klb2+klg4+klb5+klb3+klb1+lelong+kla1;	
  H[hUVWX_Z][hENDE]   = klb2+klg4+klb5+klb3+klb1+klelong;
  for(j=0;j<H[hUVWX_Z].size();j++) {
     H[h_VWX_Z][j]=H[hUVWX_Z][j];
     H[hW2][j]=H[hUVWX_Z][j];
  }  
  

  H[h___X_Z][hY1] = lg4;	
  H[h___X_Z][h_____Z] = klg4+lb5;	
  H[h___X_Z][h___XYZ] = klg4+klb5+lb3+la5+la4;	
  H[h___X_Z][h___X_Z] = klg4+klb5+lb3+la5+kla4;	
  H[h___X_Z][h___XY_] = klg4+klb5+lb3+kla5+la4;	
  H[h___X_Z][h___X__] = klg4+klb5+lb3+kla5+kla4;	
  H[h___X_Z][hUVWXYZ] = klg4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[h___X_Z][hUV_XYZ] = klg4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[h___X_Z][hUVWX_Z] = klg4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[h___X_Z][hUV_X_Z] = klg4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[h___X_Z][hUVWXY_] = klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[h___X_Z][hUV_XY_] = klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[h___X_Z][hUVWX__] = klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[h___X_Z][hUV_X__] = klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[h___X_Z][hUVW___] = klg4+klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[h___X_Z][hUV____] = klg4+klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[h___X_Z][h_VWXYZ] = klg4+klb5+klb3+lb1+la3+la5+la4+la2;	
  H[h___X_Z][h_V_XYZ] = klg4+klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[h___X_Z][h_VWX_Z] = klg4+klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[h___X_Z][h_V_X_Z] = klg4+klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[h___X_Z][h_VWXY_] = klg4+klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[h___X_Z][h_V_XY_] = klg4+klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[h___X_Z][h_VWX__] = klg4+klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[h___X_Z][h_V_X__] = klg4+klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[h___X_Z][h_VW___] = klg4+klb5+klb3+lb1+kla3+la2;	
  H[h___X_Z][h_V____] = klg4+klb5+klb3+lb1+kla3+kla2;	
  H[h___X_Z][hU_____] = klg4+klb5+klb3+klb1+lelong+kla1;	
  H[h___X_Z][hENDE]   = klg4+klb5+klb3+klb1+klelong;

  
  H[hUV_X_Z][hW2] = lg2;
  H[hUV_X_Z][hY1] = klg2+lg4;	
  H[hUV_X_Z][h_____Z] = klg2+klg4+lb5;	
  H[hUV_X_Z][h___XYZ] = klg2+klg4+klb5+lb3+la5+la4;	
  H[hUV_X_Z][h___X_Z] = klg2+klg4+klb5+lb3+la5+kla4;	
  H[hUV_X_Z][h___XY_] = klg2+klg4+klb5+lb3+kla5+la4;	
  H[hUV_X_Z][h___X__] = klg2+klg4+klb5+lb3+kla5+kla4;	
  H[hUV_X_Z][hUVWXYZ] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUV_X_Z][hUV_XYZ] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUV_X_Z][hUVWX_Z] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUV_X_Z][hUV_X_Z] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUV_X_Z][hUVWXY_] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUV_X_Z][hUV_XY_] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUV_X_Z][hUVWX__] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUV_X_Z][hUV_X__] = klg2+klg4+klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUV_X_Z][hUVW___] = klg2+klg4+klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUV_X_Z][hUV____] = klg2+klg4+klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUV_X_Z][h_VWXYZ] = klg2+klg4+klb5+klb3+lb1+la3+la5+la4+la2;	
  H[hUV_X_Z][h_V_XYZ] = klg2+klg4+klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUV_X_Z][h_VWX_Z] = klg2+klg4+klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUV_X_Z][h_V_X_Z] = klg2+klg4+klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUV_X_Z][h_VWXY_] = klg2+klg4+klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUV_X_Z][h_V_XY_] = klg2+klg4+klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUV_X_Z][h_VWX__] = klg2+klg4+klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUV_X_Z][h_V_X__] = klg2+klg4+klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUV_X_Z][h_VW___] = klg2+klg4+klb5+klb3+lb1+kla3+la2;	
  H[hUV_X_Z][h_V____] = klg2+klg4+klb5+klb3+lb1+kla3+kla2;	
  H[hUV_X_Z][hU_____] = klg2+klg4+klb5+klb3+klb1+lelong+kla1;	
  H[hUV_X_Z][hENDE]   = klg2+klg4+klb5+klb3+klb1+klelong;
  for(j=0;j<H[hUV_X_Z].size();j++)
     H[h_V_X_Z][j]=H[hUV_X_Z][j];
  
     
  H[hUVWXY_][hW3] = lb2;
  H[hUVWXY_][hY2] = klb2+lb4;	
  H[hUVWXY_][h_____Z] = klb2+klb4+lg5;	
  H[hUVWXY_][h___XYZ] = klb2+klb4+klg5+lb3+la5+la4;	
  H[hUVWXY_][h___X_Z] = klb2+klb4+klg5+lb3+la5+kla4;	
  H[hUVWXY_][h___XY_] = klb2+klb4+klg5+lb3+kla5+la4;	
  H[hUVWXY_][h___X__] = klb2+klb4+klg5+lb3+kla5+kla4;	
  H[hUVWXY_][hUVWXYZ] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUVWXY_][hUV_XYZ] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUVWXY_][hUVWX_Z] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUVWXY_][hUV_X_Z] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUVWXY_][hUVWXY_] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUVWXY_][hUV_XY_] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUVWXY_][hUVWX__] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUVWXY_][hUV_X__] = klb2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUVWXY_][hUVW___] = klb2+klb4+klg5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUVWXY_][hUV____] = klb2+klb4+klg5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUVWXY_][h_VWXYZ] = klb2+klb4+klg5+klb3+lb1+la3+la5+la4+la2;	
  H[hUVWXY_][h_V_XYZ] = klb2+klb4+klg5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUVWXY_][h_VWX_Z] = klb2+klb4+klg5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUVWXY_][h_V_X_Z] = klb2+klb4+klg5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUVWXY_][h_VWXY_] = klb2+klb4+klg5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUVWXY_][h_V_XY_] = klb2+klb4+klg5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUVWXY_][h_VWX__] = klb2+klb4+klg5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUVWXY_][h_V_X__] = klb2+klb4+klg5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUVWXY_][h_VW___] = klb2+klb4+klg5+klb3+lb1+kla3+la2;	
  H[hUVWXY_][h_V____] = klb2+klb4+klg5+klb3+lb1+kla3+kla2;	
  H[hUVWXY_][hU_____] = klb2+klb4+klg5+klb3+klb1+lelong+kla1;	
  H[hUVWXY_][hENDE]   = klb2+klb4+klg5+klb3+klb1+klelong;
  for(j=0;j<H[hUVWXY_].size();j++) {
     H[h_VWXY_][j]=H[hUVWXY_][j];
     H[hW3][j]=H[hUVWXY_][j];
  }  
  
     
  H[h___XY_][hY2] = lb4;	
  H[h___XY_][h_____Z] = klb4+lg5;	
  H[h___XY_][h___XYZ] = klb4+klg5+lb3+la5+la4;	
  H[h___XY_][h___X_Z] = klb4+klg5+lb3+la5+kla4;	
  H[h___XY_][h___XY_] = klb4+klg5+lb3+kla5+la4;	
  H[h___XY_][h___X__] = klb4+klg5+lb3+kla5+kla4;	
  H[h___XY_][hUVWXYZ] = klb4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[h___XY_][hUV_XYZ] = klb4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[h___XY_][hUVWX_Z] = klb4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[h___XY_][hUV_X_Z] = klb4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[h___XY_][hUVWXY_] = klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[h___XY_][hUV_XY_] = klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[h___XY_][hUVWX__] = klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[h___XY_][hUV_X__] = klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[h___XY_][hUVW___] = klb4+klg5+klb3+klb1+lelong+la1+kla3+la2;	
  H[h___XY_][hUV____] = klb4+klg5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[h___XY_][h_VWXYZ] = klb4+klg5+klb3+lb1+la3+la5+la4+la2;	
  H[h___XY_][h_V_XYZ] = klb4+klg5+klb3+lb1+la3+la5+la4+kla2;	
  H[h___XY_][h_VWX_Z] = klb4+klg5+klb3+lb1+la3+la5+kla4+la2;	
  H[h___XY_][h_V_X_Z] = klb4+klg5+klb3+lb1+la3+la5+kla4+kla2;	
  H[h___XY_][h_VWXY_] = klb4+klg5+klb3+lb1+la3+kla5+la4+la2;	
  H[h___XY_][h_V_XY_] = klb4+klg5+klb3+lb1+la3+kla5+la4+kla2;	
  H[h___XY_][h_VWX__] = klb4+klg5+klb3+lb1+la3+kla5+kla4+la2;	
  H[h___XY_][h_V_X__] = klb4+klg5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[h___XY_][h_VW___] = klb4+klg5+klb3+lb1+kla3+la2;	
  H[h___XY_][h_V____] = klb4+klg5+klb3+lb1+kla3+kla2;	
  H[h___XY_][hU_____] = klb4+klg5+klb3+klb1+lelong+kla1;	
  H[h___XY_][hENDE]   = klb4+klg5+klb3+klb1+klelong;
  for(j=0;j<H[h___XY_].size();j++)
     H[hY2][j]=H[h___XY_][j];
  
     
  H[hUV_XY_][hW3] = lg2;
  H[hUV_XY_][hY2] = klg2+lb4;	
  H[hUV_XY_][h_____Z] = klg2+klb4+lg5;	
  H[hUV_XY_][h___XYZ] = klg2+klb4+klg5+lb3+la5+la4;	
  H[hUV_XY_][h___X_Z] = klg2+klb4+klg5+lb3+la5+kla4;	
  H[hUV_XY_][h___XY_] = klg2+klb4+klg5+lb3+kla5+la4;	
  H[hUV_XY_][h___X__] = klg2+klb4+klg5+lb3+kla5+kla4;	
  H[hUV_XY_][hUVWXYZ] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUV_XY_][hUV_XYZ] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUV_XY_][hUVWX_Z] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUV_XY_][hUV_X_Z] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUV_XY_][hUVWXY_] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUV_XY_][hUV_XY_] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUV_XY_][hUVWX__] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUV_XY_][hUV_X__] = klg2+klb4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUV_XY_][hUVW___] = klg2+klb4+klg5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUV_XY_][hUV____] = klg2+klb4+klg5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUV_XY_][h_VWXYZ] = klg2+klb4+klg5+klb3+lb1+la3+la5+la4+la2;	
  H[hUV_XY_][h_V_XYZ] = klg2+klb4+klg5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUV_XY_][h_VWX_Z] = klg2+klb4+klg5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUV_XY_][h_V_X_Z] = klg2+klb4+klg5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUV_XY_][h_VWXY_] = klg2+klb4+klg5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUV_XY_][h_V_XY_] = klg2+klb4+klg5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUV_XY_][h_VWX__] = klg2+klb4+klg5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUV_XY_][h_V_X__] = klg2+klb4+klg5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUV_XY_][h_VW___] = klg2+klb4+klg5+klb3+lb1+kla3+la2;	
  H[hUV_XY_][h_V____] = klg2+klb4+klg5+klb3+lb1+kla3+kla2;	
  H[hUV_XY_][hU_____] = klg2+klb4+klg5+klb3+klb1+lelong+kla1;	
  H[hUV_XY_][hENDE]   = klg2+klb4+klg5+klb3+klb1+klelong;
  for(j=0;j<H[hUV_XY_].size();j++)
     H[h_V_XY_][j]=H[hUV_XY_][j];
  
     
  H[hUVWX__][hW4] = lb2;
  H[hUVWX__][hY2] = klb2+lg4;	
  H[hUVWX__][h_____Z] = klb2+klg4+lg5;	
  H[hUVWX__][h___XYZ] = klb2+klg4+klg5+lb3+la5+la4;	
  H[hUVWX__][h___X_Z] = klb2+klg4+klg5+lb3+la5+kla4;	
  H[hUVWX__][h___XY_] = klb2+klg4+klg5+lb3+kla5+la4;	
  H[hUVWX__][h___X__] = klb2+klg4+klg5+lb3+kla5+kla4;	
  H[hUVWX__][hUVWXYZ] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUVWX__][hUV_XYZ] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUVWX__][hUVWX_Z] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUVWX__][hUV_X_Z] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUVWX__][hUVWXY_] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUVWX__][hUV_XY_] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUVWX__][hUVWX__] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUVWX__][hUV_X__] = klb2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUVWX__][hUVW___] = klb2+klg4+klg5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUVWX__][hUV____] = klb2+klg4+klg5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUVWX__][h_VWXYZ] = klb2+klg4+klg5+klb3+lb1+la3+la5+la4+la2;	
  H[hUVWX__][h_V_XYZ] = klb2+klg4+klg5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUVWX__][h_VWX_Z] = klb2+klg4+klg5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUVWX__][h_V_X_Z] = klb2+klg4+klg5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUVWX__][h_VWXY_] = klb2+klg4+klg5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUVWX__][h_V_XY_] = klb2+klg4+klg5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUVWX__][h_VWX__] = klb2+klg4+klg5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUVWX__][h_V_X__] = klb2+klg4+klg5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUVWX__][h_VW___] = klb2+klg4+klg5+klb3+lb1+kla3+la2;	
  H[hUVWX__][h_V____] = klb2+klg4+klg5+klb3+lb1+kla3+kla2;	
  H[hUVWX__][hU_____] = klb2+klg4+klg5+klb3+klb1+lelong+kla1;	
  H[hUVWX__][hENDE]   = klb2+klg4+klg5+klb3+klb1+klelong;
  for(j=0;j<H[hUVWX__].size();j++) {
     H[h_VWX__][j]=H[hUVWX__][j];
     H[hW4][j]=H[hUVWX__][j];
  }  
  
     
  H[h___X__][hY2] = lg4;	
  H[h___X__][h_____Z] = klg4+lg5;	
  H[h___X__][h___XYZ] = klg4+klg5+lb3+la5+la4;	
  H[h___X__][h___X_Z] = klg4+klg5+lb3+la5+kla4;	
  H[h___X__][h___XY_] = klg4+klg5+lb3+kla5+la4;	
  H[h___X__][h___X__] = klg4+klg5+lb3+kla5+kla4;	
  H[h___X__][hUVWXYZ] = klg4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[h___X__][hUV_XYZ] = klg4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[h___X__][hUVWX_Z] = klg4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[h___X__][hUV_X_Z] = klg4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[h___X__][hUVWXY_] = klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[h___X__][hUV_XY_] = klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[h___X__][hUVWX__] = klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[h___X__][hUV_X__] = klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[h___X__][hUVW___] = klg4+klg5+klb3+klb1+lelong+la1+kla3+la2;	
  H[h___X__][hUV____] = klg4+klg5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[h___X__][h_VWXYZ] = klg4+klg5+klb3+lb1+la3+la5+la4+la2;	
  H[h___X__][h_V_XYZ] = klg4+klg5+klb3+lb1+la3+la5+la4+kla2;	
  H[h___X__][h_VWX_Z] = klg4+klg5+klb3+lb1+la3+la5+kla4+la2;	
  H[h___X__][h_V_X_Z] = klg4+klg5+klb3+lb1+la3+la5+kla4+kla2;	
  H[h___X__][h_VWXY_] = klg4+klg5+klb3+lb1+la3+kla5+la4+la2;	
  H[h___X__][h_V_XY_] = klg4+klg5+klb3+lb1+la3+kla5+la4+kla2;	
  H[h___X__][h_VWX__] = klg4+klg5+klb3+lb1+la3+kla5+kla4+la2;	
  H[h___X__][h_V_X__] = klg4+klg5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[h___X__][h_VW___] = klg4+klg5+klb3+lb1+kla3+la2;	
  H[h___X__][h_V____] = klg4+klg5+klb3+lb1+kla3+kla2;	
  H[h___X__][hU_____] = klg4+klg5+klb3+klb1+lelong+kla1;	
  H[h___X__][hENDE]   = klg4+klg5+klb3+klb1+klelong;
  
     
  H[h_____Z][h_____Z] = lb5;	
  H[h_____Z][h___XYZ] = klb5+lb3+la5+la4;	
  H[h_____Z][h___X_Z] = klb5+lb3+la5+kla4;	
  H[h_____Z][h___XY_] = klb5+lb3+kla5+la4;	
  H[h_____Z][h___X__] = klb5+lb3+kla5+kla4;	
  H[h_____Z][hUVWXYZ] = klb5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[h_____Z][hUV_XYZ] = klb5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[h_____Z][hUVWX_Z] = klb5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[h_____Z][hUV_X_Z] = klb5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[h_____Z][hUVWXY_] = klb5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[h_____Z][hUV_XY_] = klb5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[h_____Z][hUVWX__] = klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[h_____Z][hUV_X__] = klb5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[h_____Z][hUVW___] = klb5+klb3+klb1+lelong+la1+kla3+la2;	
  H[h_____Z][hUV____] = klb5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[h_____Z][h_VWXYZ] = klb5+klb3+lb1+la3+la5+la4+la2;	
  H[h_____Z][h_V_XYZ] = klb5+klb3+lb1+la3+la5+la4+kla2;	
  H[h_____Z][h_VWX_Z] = klb5+klb3+lb1+la3+la5+kla4+la2;	
  H[h_____Z][h_V_X_Z] = klb5+klb3+lb1+la3+la5+kla4+kla2;	
  H[h_____Z][h_VWXY_] = klb5+klb3+lb1+la3+kla5+la4+la2;	
  H[h_____Z][h_V_XY_] = klb5+klb3+lb1+la3+kla5+la4+kla2;	
  H[h_____Z][h_VWX__] = klb5+klb3+lb1+la3+kla5+kla4+la2;	
  H[h_____Z][h_V_X__] = klb5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[h_____Z][h_VW___] = klb5+klb3+lb1+kla3+la2;	
  H[h_____Z][h_V____] = klb5+klb3+lb1+kla3+kla2;	
  H[h_____Z][hU_____] = klb5+klb3+klb1+lelong+kla1;	
  H[h_____Z][hENDE]   = klb5+klb3+klb1+klelong;

  
  H[hUV_X__][hW4] = lg2;
  H[hUV_X__][hY2] = klg2+lg4;	
  H[hUV_X__][h_____Z] = klg2+klg4+lg5;	
  H[hUV_X__][h___XYZ] = klg2+klg4+klg5+lb3+la5+la4;	
  H[hUV_X__][h___X_Z] = klg2+klg4+klg5+lb3+la5+kla4;	
  H[hUV_X__][h___XY_] = klg2+klg4+klg5+lb3+kla5+la4;	
  H[hUV_X__][h___X__] = klg2+klg4+klg5+lb3+kla5+kla4;	
  H[hUV_X__][hUVWXYZ] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUV_X__][hUV_XYZ] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUV_X__][hUVWX_Z] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUV_X__][hUV_X_Z] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUV_X__][hUVWXY_] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUV_X__][hUV_XY_] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUV_X__][hUVWX__] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUV_X__][hUV_X__] = klg2+klg4+klg5+klb3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUV_X__][hUVW___] = klg2+klg4+klg5+klb3+klb1+lelong+la1+kla3+la2;	
  H[hUV_X__][hUV____] = klg2+klg4+klg5+klb3+klb1+lelong+la1+kla3+kla2;	
  H[hUV_X__][h_VWXYZ] = klg2+klg4+klg5+klb3+lb1+la3+la5+la4+la2;	
  H[hUV_X__][h_V_XYZ] = klg2+klg4+klg5+klb3+lb1+la3+la5+la4+kla2;	
  H[hUV_X__][h_VWX_Z] = klg2+klg4+klg5+klb3+lb1+la3+la5+kla4+la2;	
  H[hUV_X__][h_V_X_Z] = klg2+klg4+klg5+klb3+lb1+la3+la5+kla4+kla2;	
  H[hUV_X__][h_VWXY_] = klg2+klg4+klg5+klb3+lb1+la3+kla5+la4+la2;	
  H[hUV_X__][h_V_XY_] = klg2+klg4+klg5+klb3+lb1+la3+kla5+la4+kla2;	
  H[hUV_X__][h_VWX__] = klg2+klg4+klg5+klb3+lb1+la3+kla5+kla4+la2;	
  H[hUV_X__][h_V_X__] = klg2+klg4+klg5+klb3+lb1+la3+kla5+kla4+kla2;	
  H[hUV_X__][h_VW___] = klg2+klg4+klg5+klb3+lb1+kla3+la2;	
  H[hUV_X__][h_V____] = klg2+klg4+klg5+klb3+lb1+kla3+kla2;	
  H[hUV_X__][hU_____] = klg2+klg4+klg5+klb3+klb1+lelong+kla1;	
  H[hUV_X__][hENDE]   = klg2+klg4+klg5+klb3+klb1+klelong;
  for(j=0;j<H[hUV_X__].size();j++)
     H[h_V_X__][j]=H[hUV_X__][j];
  
  
  H[hUVW___][hW5] = lb2;
  H[hUVW___][h___XYZ] = klb2+lg3+la5+la4;	
  H[hUVW___][h___X_Z] = klb2+lg3+la5+kla4;	
  H[hUVW___][h___XY_] = klb2+lg3+kla5+la4;	
  H[hUVW___][h___X__] = klb2+lg3+kla5+kla4;	
  H[hUVW___][hUVWXYZ] = klb2+klg3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUVW___][hUV_XYZ] = klb2+klg3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUVW___][hUVWX_Z] = klb2+klg3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUVW___][hUV_X_Z] = klb2+klg3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUVW___][hUVWXY_] = klb2+klg3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUVW___][hUV_XY_] = klb2+klg3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUVW___][hUVWX__] = klb2+klg3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUVW___][hUV_X__] = klb2+klg3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUVW___][hUVW___] = klb2+klg3+klb1+lelong+la1+kla3+la2;	
  H[hUVW___][hUV____] = klb2+klg3+klb1+lelong+la1+kla3+kla2;	
  H[hUVW___][h_VWXYZ] = klb2+klg3+lb1+la3+la5+la4+la2;	
  H[hUVW___][h_V_XYZ] = klb2+klg3+lb1+la3+la5+la4+kla2;	
  H[hUVW___][h_VWX_Z] = klb2+klg3+lb1+la3+la5+kla4+la2;	
  H[hUVW___][h_V_X_Z] = klb2+klg3+lb1+la3+la5+kla4+kla2;	
  H[hUVW___][h_VWXY_] = klb2+klg3+lb1+la3+kla5+la4+la2;	
  H[hUVW___][h_V_XY_] = klb2+klg3+lb1+la3+kla5+la4+kla2;	
  H[hUVW___][h_VWX__] = klb2+klg3+lb1+la3+kla5+kla4+la2;	
  H[hUVW___][h_V_X__] = klb2+klg3+lb1+la3+kla5+kla4+kla2;	
  H[hUVW___][h_VW___] = klb2+klg3+lb1+kla3+la2;	
  H[hUVW___][h_V____] = klb2+klg3+lb1+kla3+kla2;	
  H[hUVW___][hU_____] = klb2+klg3+klb1+lelong+kla1;	
  H[hUVW___][hENDE]   = klb2+klg3+klb1+klelong;
  for(j=0;j<H[hUVW___].size();j++) {
     H[h_VW___][j]=H[hUVW___][j];
     H[hW5][j]=H[hUVW___][j];
  }

  
  H[hUV____][hW5] = lg2;
  H[hUV____][h___XYZ] = klg2+lg3+la5+la4;	
  H[hUV____][h___X_Z] = klg2+lg3+la5+kla4;	
  H[hUV____][h___XY_] = klg2+lg3+kla5+la4;	
  H[hUV____][h___X__] = klg2+lg3+kla5+kla4;	
  H[hUV____][hUVWXYZ] = klg2+klg3+klb1+lelong+la1+la3+la5+la4+la2;	
  H[hUV____][hUV_XYZ] = klg2+klg3+klb1+lelong+la1+la3+la5+la4+kla2;	
  H[hUV____][hUVWX_Z] = klg2+klg3+klb1+lelong+la1+la3+la5+kla4+la2;	
  H[hUV____][hUV_X_Z] = klg2+klg3+klb1+lelong+la1+la3+la5+kla4+kla2;	
  H[hUV____][hUVWXY_] = klg2+klg3+klb1+lelong+la1+la3+kla5+la4+la2;	
  H[hUV____][hUV_XY_] = klg2+klg3+klb1+lelong+la1+la3+kla5+la4+kla2;	
  H[hUV____][hUVWX__] = klg2+klg3+klb1+lelong+la1+la3+kla5+kla4+la2;	
  H[hUV____][hUV_X__] = klg2+klg3+klb1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hUV____][hUVW___] = klg2+klg3+klb1+lelong+la1+kla3+la2;	
  H[hUV____][hUV____] = klg2+klg3+klb1+lelong+la1+kla3+kla2;	
  H[hUV____][h_VWXYZ] = klg2+klg3+lb1+la3+la5+la4+la2;	
  H[hUV____][h_V_XYZ] = klg2+klg3+lb1+la3+la5+la4+kla2;	
  H[hUV____][h_VWX_Z] = klg2+klg3+lb1+la3+la5+kla4+la2;	
  H[hUV____][h_V_X_Z] = klg2+klg3+lb1+la3+la5+kla4+kla2;	
  H[hUV____][h_VWXY_] = klg2+klg3+lb1+la3+kla5+la4+la2;	
  H[hUV____][h_V_XY_] = klg2+klg3+lb1+la3+kla5+la4+kla2;	
  H[hUV____][h_VWX__] = klg2+klg3+lb1+la3+kla5+kla4+la2;	
  H[hUV____][h_V_X__] = klg2+klg3+lb1+la3+kla5+kla4+kla2;	
  H[hUV____][h_VW___] = klg2+klg3+lb1+kla3+la2;	
  H[hUV____][h_V____] = klg2+klg3+lb1+kla3+kla2;	
  H[hUV____][hU_____] = klg2+klg3+klb1+lelong+kla1;	
  H[hUV____][hENDE]   = klg2+klg3+klb1+klelong;
  for(j=0;j<H[hUV____].size();j++)
     H[h_V____][j]=H[hUV____][j];


  H[hU_____][hUVWXYZ] = klg1+lelong+la1+la3+la5+la4+la2;	
  H[hU_____][hUV_XYZ] = klg1+lelong+la1+la3+la5+la4+kla2;	
  H[hU_____][hUVWX_Z] = klg1+lelong+la1+la3+la5+kla4+la2;	
  H[hU_____][hUV_X_Z] = klg1+lelong+la1+la3+la5+kla4+kla2;	
  H[hU_____][hUVWXY_] = klg1+lelong+la1+la3+kla5+la4+la2;	
  H[hU_____][hUV_XY_] = klg1+lelong+la1+la3+kla5+la4+kla2;	
  H[hU_____][hUVWX__] = klg1+lelong+la1+la3+kla5+kla4+la2;	
  H[hU_____][hUV_X__] = klg1+lelong+la1+la3+kla5+kla4+kla2;	
  H[hU_____][hUVW___] = klg1+lelong+la1+kla3+la2;	
  H[hU_____][hUV____] = klg1+lelong+la1+kla3+kla2;	
  H[hU_____][h_VWXYZ] = lg1+la3+la5+la4+la2;	
  H[hU_____][h_V_XYZ] = lg1+la3+la5+la4+kla2;	
  H[hU_____][h_VWX_Z] = lg1+la3+la5+kla4+la2;	
  H[hU_____][h_V_X_Z] = lg1+la3+la5+kla4+kla2;	
  H[hU_____][h_VWXY_] = lg1+la3+kla5+la4+la2;	
  H[hU_____][h_V_XY_] = lg1+la3+kla5+la4+kla2;	
  H[hU_____][h_VWX__] = lg1+la3+kla5+kla4+la2;	
  H[hU_____][h_V_X__] = lg1+la3+kla5+kla4+kla2;	
  H[hU_____][h_VW___] = lg1+kla3+la2;	
  H[hU_____][h_V____] = lg1+kla3+kla2;	
  H[hU_____][hU_____] = klg1+lelong+kla1;	
  H[hU_____][hENDE]   = klg1+klelong;

  if(model==TKF2) {
     for(int i=0;i<(int)H.size();++i)
       if(i!=hSTART)
         for(int j=0;j<(int)H[i].size();++j) {
	     if(H[i][j]>-DBL_MAX)
	        H[i][j]+=klr;
	     if(i==j)
	        H[i][j]=untidy_addlog(log(r),H[i][j]);
	 } 
  }
}


double indel::retprob(zustand z1, zustand z2)
{
  if(z1==NOSTATE || z2==NOSTATE || z1==ENDE)
  {
     cerr << "At least one of the states that was passed\n"
          << "to retprob is out of range.\n";
     exit(EXIT_FAILURE);
  }
  
  return exp(P[z1][z2]);
}

double indel::retlnprob(int z1, int z2)
{
  if(z1==NOSTATE || z2==NOSTATE || z1==ENDE)
  {
     cerr << "At least one of the states that was passed\n"
          << "to retprob is out of range.\n";
     exit(EXIT_FAILURE);
  }
  
  return P[z1][z2];
}

zustand indel::nextpair(zustand now, bool stillseq)
{
   vector<zustand> possible;
   vector<double> probs;
   if(now==NOSTATE || now==ENDE)
   {
     cerr << "Unreasonable state " <<(int)now 
	  <<" in function nextpair\n";
     exit(EXIT_FAILURE);
   }
   if(stillseq) {
      possible.push_back(INSERTION);
      probs.push_back(exp(P[now][INSERTION]));
      possible.push_back(NOSTATE);
      probs.push_back(1-exp(P[now][INSERTION]));
   } else {
      possible.push_back(INSERTION);
      probs.push_back(exp(P[now][INSERTION]));
      possible.push_back(ENDE);
      probs.push_back(1-exp(P[now][INSERTION]));
   }

  // Zustand auswaehlen
  int i=randomchoice(probs);
  if(possible[i]==NOSTATE) { // noch zwischen DELETION und PAIR entscheiden
     possible.clear();
     probs.clear();
     possible.push_back(PAIR);
     probs.push_back(exp(P[now][PAIR]));
     possible.push_back(DELETION);
     probs.push_back(exp(P[now][DELETION]));
     i=randomchoice(probs);
  }
  return possible[i];
}

triplet_zus indel::nexttriplet(triplet_zus now, zustand then, unsigned int internalcount)
{
  vector<triplet_zus> possible;
  vector<double> probs;
  int i;
   
  if(now==tNOSTATE || now==tENDE || now==t_YZ)
  {
     cerr << "Unreasonable state " <<(int)now 
	  <<" in function nexttriplet\n";
     exit(EXIT_FAILURE);
  }

  switch(then) {
          case PAIR :
             possible.push_back(tXYZ);
	     probs.push_back(exp(T[now][tXYZ]));
	     break;
          case INSERTION : // the sequence at the right branch is considered ancestral
             possible.push_back(tXY_);
	     probs.push_back(exp(T[now][tXY_]));
	     switch(now) {
	       case tSTART :
	       case tXYZ :
	       case tX_Z :
	         possible.push_back(t_Y_1);
		 probs.push_back(exp(T[now][t_Y_1]));
		 break;
	       case tXY_ :
	       case tX__ :
	         possible.push_back(t_Y_2);
		 probs.push_back(exp(T[now][t_Y_2]));
		 break;
	       default :
	         ;
	     }
             break;
	  case DELETION : // the sequence at the right branch is considered ancestral
	     possible.push_back(tX_Z);
	     probs.push_back(exp(T[now][tX_Z]));
	     possible.push_back(t__Z);
	     probs.push_back(exp(T[now][t__Z]));
	     break;
	  case START :
	     break;
	  case ENDE :
	     possible.push_back(tENDE);
	     probs.push_back(exp(T[now][tENDE]));
	     break;
          default :
             cerr << "Unreasonable state " <<(int)then 
	          <<" in function nexttriplet\n";
             exit(EXIT_FAILURE);
  }
   // Der folgende Programmteil soll verhindern, dass man aufgrund
   // moeglicher Groessenordnungsunterschiede nur noch innere Positionen
   // erzeugt
   bool exact=TRUE;
   switch(model) {
      case TKF1 :
         exact=TRUE;
         break;
      case TKF2 :
         if(params[2]>CRITICAL_R)
	    exact=FALSE;
	 else
	    exact=TRUE;
         break;
      default :
         cerr << "Unknown indel model\n";
	 exit(EXIT_FAILURE);
   }
   if(internalcount<MAXINTERNALSTRETCH) {
      if(now==tX__ && !exact) {
        if(randomunitintervall()<exp(T[tX__][tX__]))
           return tX__;
      } else {
        possible.push_back(tX__);
        probs.push_back(exp(T[now][tX__]));
      }
   }
  // Zustand auswaehlen
  i=randomchoice(probs);
  
  return possible[i];

}


quart_zus indel::nextquartet(quart_zus now, triplet_zus then, unsigned int internalcount)
{
  vector<quart_zus> possible;
  vector<double> probs;
  int i;
   
  if(now==qNOSTATE || now==qENDE)
  {
      cerr << "Unreasonable state " <<(int)now 
	   <<" in function nextquartet\n";
      exit(EXIT_FAILURE);
  }
  switch(then) {
      case tXYZ : // X im Triplett entspricht W im Quartett
	 possible.push_back(qWXYZ);
	 probs.push_back(exp(Q[now][qWXYZ]));
	 break;
      case tX_Z : // X im Triplett entspricht W im Quartett
	 possible.push_back(qWX_Z);
	 probs.push_back(exp(Q[now][qWX_Z]));
	 break;
      case tXY_ : // X im Triplett entspricht W im Quartett
	 possible.push_back(qWXY_);
	 probs.push_back(exp(Q[now][qWXY_]));
	 break;
      case tX__ : // X im Triplett entspricht W im Quartett
	 possible.push_back(qWX__);
	 probs.push_back(exp(Q[now][qWX__]));
	 possible.push_back(qW___);
	 probs.push_back(exp(Q[now][qW___]));
	 break;
      case t_YZ : // t_YZ ist nur in diesem Zusammenhang sinnvoll
	 possible.push_back(q_XYZ);
	 probs.push_back(exp(Q[now][q_XYZ]));
	 break;
      case t_Y_1 :
      case t_Y_2 :
	     switch(now) {
	         case qSTART :
		 case q_XYZ :
		 case q_X_Z :
		 case q__Y_1 :
		    possible.push_back(q__Y_1);
		    probs.push_back(exp(Q[now][q__Y_1]));
		    break;
		 case q_XY_ :
		 case q_X__ :
		 case q__Y_2 :
		   possible.push_back(q__Y_2);
		   probs.push_back(exp(Q[now][q__Y_2]));
		   break;
		 case qWXYZ :
		 case qWX_Z :
		 case q__Y_3 :
		   possible.push_back(q__Y_3);
		   probs.push_back(exp(Q[now][q__Y_3]));
		   break;
		 case qWXY_ :
		 case qWX__ :
		 case q__Y_4 :
		   possible.push_back(q__Y_4);
		   probs.push_back(exp(Q[now][q__Y_4]));
		   break;
		 case q___Z1 :
		 case q___Z2 :
		 case qW___ :
		    break;
                 default :
                    cerr << "Unreasonable state " <<(int)now 
	                 <<" in function nextquartet\n";
                    exit(EXIT_FAILURE);
             }
	possible.push_back(q_XY_);
	probs.push_back(exp(Q[now][q_XY_]));
	break;
     case t__Z :
	     switch(now) {
                 case qSTART :
		 case q_XYZ :
		 case q_X_Z :
		 case q_XY_ :
		 case q_X__ :
		 case q__Y_1 :
		 case q__Y_2 :
		 case q___Z1 :
		    possible.push_back(q___Z1);
		    probs.push_back(exp(Q[now][q___Z1]));
		    break;
		 case qWXYZ :
		 case qWX_Z :
		 case qWXY_ :
		 case qWX__ :
		 case q__Y_3 :
		 case q__Y_4 :
		 case q___Z2 :
		    possible.push_back(q___Z2);
		    probs.push_back(exp(Q[now][q___Z2]));
		    break;
		 case qW___ :
		    break;
                 default :
                    cerr << "Unreasonable state " <<(int)now 
	                 <<" in function nextquartet\n";
                    exit(EXIT_FAILURE);
             }
	possible.push_back(q_X_Z);
	probs.push_back(exp(Q[now][q_X_Z]));
	break;
     case tENDE :
	possible.push_back(qENDE);
	probs.push_back(exp(Q[now][qENDE]));
	break;
     case tSTART :
	break;
     default :
        cerr << "Unreasonable state " <<(int)then 
	     <<" in function nextquartet\n";
        exit(EXIT_FAILURE);
   }

   // Der folgende Programmteil soll verhindern, dass man aufgrund
   // moeglicher Groessenordnungsunterschiede nur noch innere Positionen
   // erzeugt
   bool exact=TRUE;
   switch(model) {
      case TKF1 :
         exact=TRUE;
         break;
      case TKF2 :
         if(params[2]>CRITICAL_R)
	    exact=FALSE;
	 else
	    exact=TRUE;
         break;
      default :
         cerr << "Unknown indel model\n";
	 exit(EXIT_FAILURE);
   }
   
   if(internalcount<MAXINTERNALSTRETCH) {
      if(now==q_X__ && !exact) {
        if(randomunitintervall()<exp(Q[q_X__][q_X__]))
           return q_X__;
      } else {
         possible.push_back(q_X__);
         probs.push_back(exp(Q[now][q_X__]));
      }
   }
  
  // Zustand auswaehlen
  i=randomchoice(probs);
  
  return possible[i];

}

hex_zus indel::nexthextet(hex_zus now, quart_zus then, unsigned int internalcount)
{
  vector<hex_zus> possible;
  vector<double> probs;
  int i;
   
  if(now==hNOSTATE || now==hENDE)
  {
      cerr << "Unreasonable state " <<(int)now 
	   <<" in function nexthextet\n";
      exit(EXIT_FAILURE);
  }
  // 'W' im Quartettzustand entspricht 'U' im Hextettzustand!
  // 'X' im Quartettzustand entspricht 'W' im Hextettzustand!

  switch (then) {
     case qWXYZ :
        possible.push_back(hUVWXYZ);
        probs.push_back(exp(H[now][hUVWXYZ]));
	break;
     case qWXY_ :
        possible.push_back(hUVWXY_);
        probs.push_back(exp(H[now][hUVWXY_]));
	break;
     case qWX_Z :
        possible.push_back(hUVWX_Z);
        probs.push_back(exp(H[now][hUVWX_Z]));
	break;
     case q_XYZ :
        possible.push_back(h_VWXYZ);
        probs.push_back(exp(H[now][h_VWXYZ]));
	break;
     case qWX__ :
        possible.push_back(hUVW___);
        probs.push_back(exp(H[now][hUVW___]));
        possible.push_back(hUVWX__);
        probs.push_back(exp(H[now][hUVWX__]));
        break;
     case q_XY_ :
        possible.push_back(h_VWXY_);
        probs.push_back(exp(H[now][h_VWXY_]));
        break;
     case q_X_Z :
        possible.push_back(h_VWX_Z);
        probs.push_back(exp(H[now][h_VWX_Z]));
        break;
     case qW_YZ :
        possible.push_back(hUV_XYZ);
        probs.push_back(exp(H[now][hUV_XYZ]));
        break;
     case qW_Y_ :
        possible.push_back(hUV_XY_);
        probs.push_back(exp(H[now][hUV_XY_]));
        break;
     case qW__Z :
        possible.push_back(hUV_X_Z);
        probs.push_back(exp(H[now][hUV_X_Z]));
        break;
     case q__YZ :
        possible.push_back(h_V_XYZ);
        probs.push_back(exp(H[now][h_V_XYZ]));
        if(now!=hU_____) {
            possible.push_back(h___XYZ);
            probs.push_back(exp(H[now][h___XYZ]));
	}
        break;
     case qW___ :
        possible.push_back(hU_____);
        probs.push_back(exp(H[now][hU_____]));
        possible.push_back(hUV____);
        probs.push_back(exp(H[now][hUV____]));
        possible.push_back(hUV_X__);
        probs.push_back(exp(H[now][hUV_X__]));
        break;
     case q_X__ :
        possible.push_back(h_VW___);
        probs.push_back(exp(H[now][h_VW___]));
        possible.push_back(h_VWX__);
        probs.push_back(exp(H[now][h_VWX__]));
        switch(now) {
	   case hSTART :
	   case hUVWXYZ :
	   case h_VWXYZ :
	   case hUV_XYZ :
	   case h_V_XYZ :
	   case hW1 :
              possible.push_back(hW1);
              probs.push_back(exp(H[now][hW1]));
	      break;
	   case hUVWX_Z :
	   case h_VWX_Z :
	   case hUV_X_Z :
	   case h_V_X_Z :
	   case hW2 :
              possible.push_back(hW2);
              probs.push_back(exp(H[now][hW2]));
	      break;
	   case hUVWXY_ :
	   case h_VWXY_ :
	   case hUV_XY_ :
	   case h_V_XY_ :
	   case hW3 :
              possible.push_back(hW3);
              probs.push_back(exp(H[now][hW3]));
	      break;
	   case hUVWX__ :
	   case h_VWX__ :
	   case hUV_X__ :
	   case h_V_X__ :
	   case hW4 :
              possible.push_back(hW4);
              probs.push_back(exp(H[now][hW4]));
	      break;
	   case hUVW___ :
	   case h_VW___ :
	   case hUV____ :
	   case h_V____ :
	   case hW5 :
              possible.push_back(hW5);
              probs.push_back(exp(H[now][hW5]));
	      break;
	   case h___XYZ :
	   case h___X_Z :
	   case h___XY_ :
	   case h___X__ :
	   case hY1 :
	   case hY2 :
	   case h_____Z :
	   case hU_____ :
	      break;
           default :
              cerr << "Unreasonable state " <<(int)now 
	           <<" in function nexthextet\n";
              exit(EXIT_FAILURE);
	}
	break;
     case q__Y_1 :
     case q__Y_2 :
     case q__Y_3 :
     case q__Y_4 :
        possible.push_back(h_V_XY_);
        probs.push_back(exp(H[now][h_V_XY_]));
        switch (now) {
	   case hSTART :
	   case hUVWXYZ :
	   case h_VWXYZ :
	   case hUV_XYZ :
	   case h_V_XYZ :
	   case hW1 :
	   case hUVWX_Z :
	   case h_VWX_Z :
	   case hUV_X_Z :
	   case h_V_X_Z :
	   case hW2 :
	   case h___XYZ :
	   case h___X_Z :
	   case hY1 :
              possible.push_back(h___XY_);
              probs.push_back(exp(H[now][h___XY_]));
              possible.push_back(hY1);
              probs.push_back(exp(H[now][hY1]));
	      break;
	   case hUVWXY_ :
	   case h_VWXY_ :
	   case hUV_XY_ :
	   case h_V_XY_ :
	   case hW3 :
	   case hUVWX__ :
	   case h_VWX__ :
	   case hUV_X__ :
	   case h_V_X__ :
	   case hW4 :
	   case h___XY_ :
	   case h___X__ :
	   case hY2 :
              possible.push_back(h___XY_);
              probs.push_back(exp(H[now][h___XY_]));
              possible.push_back(hY2);
              probs.push_back(exp(H[now][hY2]));
	      break;
	   case hUVW___ :
	   case h_VW___ :
	   case hUV____ :
	   case h_V____ :
	   case hW5 :
           case h_____Z :
              possible.push_back(h___XY_);
              probs.push_back(exp(H[now][h___XY_]));
	      break;
	   case hU_____ :
	      break;
           default :
              cerr << "Unreasonable state " <<(int)now 
	           <<" in function nexthextet\n";
              exit(EXIT_FAILURE);
	}
	break;
     case q___Z1 :
     case q___Z2 :
        possible.push_back(h_V_X_Z);
        probs.push_back(exp(H[now][h_V_X_Z]));
        if(now!=hU_____) {
            possible.push_back(h___X_Z);
            probs.push_back(exp(H[now][h___X_Z]));
	    if(now!=hUVW___ && now!=h_VW___ && now!=hUV____ && now!=h_V____ && now!=hW5) {
                possible.push_back(h_____Z);
                probs.push_back(exp(H[now][h_____Z]));
	    }
	}
        break;
     case qENDE :
        possible.push_back(hENDE);
        probs.push_back(exp(H[now][hENDE]));
	break;
     case qSTART :
        break;
     default :
        cerr << "Unreasonable state " <<(int)then 
	     <<" in function nexthextet\n";
        exit(EXIT_FAILURE);
   }

   // Der folgende Programmteil soll verhindern, dass man aufgrund
   // moeglicher Groessenordnungsunterschiede nur noch innere Positionen
   // erzeugt
   bool exact=TRUE;
   switch(model) {
      case TKF1 :
         exact=TRUE;
         break;
      case TKF2 :
         if(params[2]>CRITICAL_R)
	    exact=FALSE;
	 else
	    exact=TRUE;
         break;
      default :
         cerr << "Unknown indel model\n";
	 exit(EXIT_FAILURE);
   }
   
   if(internalcount<MAXINTERNALSTRETCH) {
      if(now==h___X__ && !exact) {
        if(randomunitintervall()<exp(H[h___X__][h___X__]))
           return h___X__;
      } else {
         if(now!=hU_____) {
            possible.push_back(h___X__);
            probs.push_back(exp(H[now][h___X__]));
         }
      }
      if(now==h_V_X__ && !exact) {
        if(randomunitintervall()< exp(H[h_V_X__][h_V_X__]))
           return h_V_X__;
      } else {
          possible.push_back(h_V_X__);
          probs.push_back(exp(H[now][h_V_X__]));
      }   
      if(now==h_V____ && !exact) {
         if(randomunitintervall()<exp(H[h_V____][h_V____]))
            return h_V____;
      } else {
         possible.push_back(h_V____);
         probs.push_back(exp(H[now][h_V____]));
      }
   }
  // Zustand auswaehlen
  i=randomchoice(probs);
  
  return possible[i];

}
   
   
indel indel::operator=(indel id2)
{
   params=id2.params;
   model=id2.model;
   pt=id2.pt;
   ty=id2.ty;
   tz=id2.tz;
   qw=id2.qw;
   qy=id2.qy;
   qz=id2.qz;
   huv=id2.huv;
   hvw=id2.hvw;
   hvx=id2.hvx;
   hxy=id2.hxy;
   hxz=id2.hxz;
   P=id2.P;
   T=id2.T;
   Q=id2.Q;
   H=id2.H;

   return *this;
}


/* unsinnig

fbindel::fbindel()
{
   cerr << "Unsinniger Konstruktor fuer fbindel\n";
   exit(EXIT_FAILURE);
}


fbindel::fbindel(indel& id)
{
   vorwaerts.resize(7);
   rueckwaerts.resize(7);
   naht.resize(7);
   for(int i=0;i<7;++i){
       vorwaerts[i].resize(7);
       rueckwaerts[i].resize(7);
       naht[i].resize(7);
       for(int j=0;j<7;++j) {
           vorwaerts[i][j]=-DBL_MAX;
	   rueckwaerts[i][j]=-DBL_MAX;
           naht[i][j]=-DBL_MAX;
       }
   }
   double alpha, beta, gamma, lambda, mue, r=0.0;
   vector<double> params;
   switch(id.returnparams(params)) {
      case TKF2 :
         r = params[2];
      case TKF1 :
         lambda = params[0];
	 mue = params[1];
	 break;
      default :
         cerr << "Unknown indel model in constructor fbindel(indel& )\n";
	 exit(EXIT_FAILURE);
   }
   alpha=id.tkf1_alpha(mue,id.pt);
   beta=id.tkf1_beta(lambda,mue,id.pt);
   gamma=id.tkf1_gamma(lambda,mue,id.pt);

   vorwaerts[fbSTART][fbXY]=id.retlnprob(START,PAIR);
   vorwaerts[fbSTART][fbX_]=id.retlnprob(START,DELETION);
   vorwaerts[fbSTART][fbY1]=id.retlnprob(START,INSERTION);
   vorwaerts[fbSTART][fbENDE]=id.retlnprob(START,ENDE);
   vorwaerts[fbXY][fbXY]=id.retlnprob(PAIR,PAIR);
   vorwaerts[fbXY][fbX_]=id.retlnprob(PAIR,DELETION);
   vorwaerts[fbXY][fbY2]=id.retlnprob(PAIR,INSERTION);
   vorwaerts[fbXY][fbENDE]=id.retlnprob(PAIR,ENDE);
   vorwaerts[fbX_][fbXY]=id.retlnprob(DELETION,PAIR);
   vorwaerts[fbX_][fbX_]=id.retlnprob(DELETION,DELETION);
   vorwaerts[fbX_][fbY3]=id.retlnprob(DELETION,INSERTION);
   vorwaerts[fbX_][fbENDE]=id.retlnprob(DELETION,ENDE);
   vorwaerts[fbY1][fbXY]=id.retlnprob(INSERTION,PAIR);
   vorwaerts[fbY1][fbX_]=id.retlnprob(INSERTION,DELETION);
   vorwaerts[fbY1][fbY1]=id.retlnprob(INSERTION,INSERTION);
   vorwaerts[fbY1][fbENDE]=id.retlnprob(INSERTION,ENDE);
   vorwaerts[fbY2][fbXY]=id.retlnprob(INSERTION,PAIR);
   vorwaerts[fbY2][fbX_]=id.retlnprob(INSERTION,DELETION);
   vorwaerts[fbY2][fbY2]=id.retlnprob(INSERTION,INSERTION);
   vorwaerts[fbY2][fbENDE]=id.retlnprob(INSERTION,ENDE);
   vorwaerts[fbY3][fbXY]=id.retlnprob(INSERTION,PAIR);
   vorwaerts[fbY3][fbX_]=id.retlnprob(INSERTION,DELETION);
   vorwaerts[fbY3][fbY3]=id.retlnprob(INSERTION,INSERTION);
   vorwaerts[fbY3][fbENDE]=id.retlnprob(INSERTION,ENDE);


   for(int i=0; i<7;++i)
      if(i!=(int)fbENDE) {
         naht[i][fbXY]=log(1.0-r)+log(1.0-lambda/mue)-log(lambda)+log(mue)-log(alpha);
         naht[i][fbX_]=log(1.0-r)+log(1.0-lambda/mue)-log(lambda)+log(mue)-log(1.0-alpha);
	 if(i==(int)fbSTART || i==(int)fbY1)
            naht[i][fbY1]=log(1.0-r)+log(1.0-beta)-log(beta);
	 if(i==(int)fbXY || i==(int)fbY2)
	    naht[i][fbY2]=log(1.0-r)+log(1.0-beta)+log(1.0-lambda/mue)-log(lambda)+log(mue)-log(alpha)-log(beta);
	 if(i==(int)fbX_ || i==(int)fbY3)
	    naht[i][fbY3]=log(1.0-r)+log(1.0-beta)+log(1.0-lambda/mue)-log(lambda)+log(mue)-log(1.0-alpha)-log(gamma);
	 naht[i][fbENDE]=0.0;
	 for(int j=0;j<7;++j)
	    if(naht[i][j]>-DBL_MAX){
	       if(vorwaerts[i][j]>-DBL_MAX)
	          naht[i][j]+=vorwaerts[i][j];
               else {
                  cerr << "Logical error in fbindel\n";
		  exit(EXIT_FAILURE);
	       }
	    }
   }


   rueckwaerts[fbENDE][fbY1]=log(1.0-lambda/mue)+log(beta);
   rueckwaerts[fbENDE][fbSTART]=log(1.0-lambda/mue)+log(1.0-beta);
   rueckwaerts[fbENDE][fbY2]=log(lambda)-log(mue)+log(alpha)+log(beta);
   rueckwaerts[fbENDE][fbXY]=log(lambda)-log(mue)+log(alpha)+log(1.0-beta);
   rueckwaerts[fbENDE][fbY3]=log(lambda)-log(mue)+log(1.0-alpha)+log(gamma);
   rueckwaerts[fbENDE][fbXY]=log(lambda)-log(mue)+log(1.0-alpha)+log(1.0-gamma);
   rueckwaerts[fbY1][fbY1]=untidy_addlog(log(r),log(1.0-r)+log(beta));
   rueckwaerts[fbY1][fbSTART]=log(1.0-r)+log(1.0-beta);
   rueckwaerts[fbY2][fbY2]=untidy_addlog(log(r),log(1.0-r)+log(beta));
   rueckwaerts[fbY2][fbXY]=log(1.0-r)+log(1.0-beta);
   rueckwaerts[fbY3][fbY3]=untidy_addlog(log(r),log(1.0-r)+log(beta));
   rueckwaerts[fbY3][fbX_]=log(1.0-r)+log(1.0-beta);
   rueckwaerts[fbXY][fbXY]=untidy_addlog(log(r),log(1.0-r)+log(lambda)-log(mue)+log(alpha)+log(1.0-beta));
   rueckwaerts[fbXY][fbX_]=log(1.0-r)+log(lambda)-log(mue)+log(1.0-alpha)+log(1.0-gamma);
   rueckwaerts[fbXY][fbY1]=log(1.0-r)+log(1.0-lambda/mue)+log(beta);
   rueckwaerts[fbXY][fbY2]=log(1.0-r)+log(lambda)-log(mue)+log(alpha)+log(beta);
   rueckwaerts[fbXY][fbY3]=log(1.0-r)+log(lambda)-log(mue)+log(1.0-alpha)+log(gamma);
   rueckwaerts[fbXY][fbSTART]=log(1.0-r)+log(1.0-lambda/mue)+log(1.0-beta);
   rueckwaerts[fbX_][fbX_]=untidy_addlog(log(r),log(1.0-r)+log(lambda)-log(mue)+log(1.0-alpha)+log(1.0-gamma));
   rueckwaerts[fbX_][fbXY]=log(1.0-r)+log(lambda)-log(mue)+log(alpha)+log(1.0-beta);
   rueckwaerts[fbX_][fbY1]=log(1.0-r)+log(1.0-lambda/mue)+log(beta);
   rueckwaerts[fbX_][fbY2]=log(1.0-r)+log(lambda)-log(mue)+log(alpha)+log(beta);
   rueckwaerts[fbX_][fbY3]=log(1.0-r)+log(lambda)-log(mue)+log(1.0-alpha)+log(gamma);
   rueckwaerts[fbX_][fbSTART]=log(1.0-r)+log(1.0-lambda/mue)+log(1.0-beta);

}


fbindel::~fbindel()
{
}

double fbindel::lnforwprob(fbzustand f1, fbzustand f2)
{
   return vorwaerts[f1][f2];
}

double fbindel::lnbackwprob(fbzustand f1, fbzustand f2)
{
   return rueckwaerts[f1][f2];
}


double fbindel::lncorrfac(fbzustand v, fbzustand r)
{
   return naht[v][r];
}

*/

