/****************************************************************************
 * This file is part of the program DOAL                                    *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <cfloat>

using namespace std;

#include "random.hh"

#define DEBUG

/******************************************************************************/
/* random numbers generator  (Numerical recipes)                              */
/******************************************************************************/

/* definitions */
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* variable */
long idum;


double randomunitintervall()
/* Long period (> 2e18) random number generator. Returns a uniform random
   deviate between 0.0 and 1.0 (exclusive of endpoint values).

   Source:
   Press et al., "Numerical recipes in C", Cambridge University Press, 1992
   (chapter 7 "Random numbers", ran2 random number generator) */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (idum <= 0) {
		if (-(idum) < 1)
			idum=1;
		else
			idum=-(idum);
		idum2=(idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(idum)/IQ1;
			idum=IA1*(idum-k*IQ1)-k*IR1;
			if (idum < 0)
				idum += IM1;
			if (j < NTAB)
				iv[j] = idum;
		}
		iy=iv[0];
	}
	k=(idum)/IQ1;
	idum=IA1*(idum-k*IQ1)-k*IR1;
	if (idum < 0)
		idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0)
		idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = idum;
	if (iy < 1)
		iy += IMM1;
	if ((temp=AM*iy) > RNMX)
		return RNMX;
	else
		return temp;
}

#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

void initrandom()
{
#ifdef DEBUG
        unsigned int Saat=(unsigned) time(NULL);
	ofstream randseed("Zufallszahl",ios::app);
	randseed<<Saat<<endl;
	randseed.close();
	srand(Saat);
#else
	srand((unsigned) time(NULL));
#endif
	idum=-(long) rand();
}

/* returns a random integer in the range [0; n - 1] */
int randominteger(int n)
{
	return (int) floor(randomunitintervall()*n);
}

void rndorder(vector<int>& zuford,unsigned int size) // zufaellige Ordnung der Zahlen 0 bis size-1
{
   zuford.resize(size);
   chooser((int)size,(int)size,zuford);
}

/* Draw s numbers from the set 0,1,2,..,t-1 and put them
   into slist (every number can be drawn only one time) */
void chooser(int t, int s, vector<int>& slist)
{
	int i, j, k, l;
	vector<bool> isfree(t);

	for (i = 0; i < t; i++) isfree[i] = TRUE;
	for (i = 0; i < s; i++) {
		/* random integer in the range [0, t-1-i] */
		j = randominteger(t-i);
		k = -1;
		l = -1;
		do {
			k++;
			if (isfree[k] == TRUE) l++; 
		} while ( l != j);
		slist[i] = k;
		isfree[k] = FALSE;
	}
}

int randomchoice(vector<int> &clust)  // zufaelliger Eintrag aus clust (Eintraege von clust sind absolute Haeufigkeiten
{
   int stateidx;
   unsigned int i;
   double R, clustsum;

   if(clust.size()==0) {
      cerr <<"randomchoice(vector<int> &clust) cannot find\n"
           <<"a random member of an empty vector.\n";
      exit(EXIT_FAILURE);
   }

   clustsum = 0.0;
   for(i=0;i<clust.size();i++)
     if(clust[i]>0)
       clustsum += (double)clust[i];
     
   R = clustsum*randomunitintervall();
   stateidx = -1;
   clustsum = 0.0;
   do {
     stateidx++;
     if(stateidx>=(int)clust.size()) {
         cerr << "Error in funnction baum::randomchoice\n"
              << "variable stateidx out of range\n";
         exit(EXIT_FAILURE);
     }
     if(clust[stateidx]>0)
       clustsum+=(double)clust[stateidx];
   } while(clustsum<R-DBL_MIN);

   return stateidx;
}

int randomchoice(vector<double> &probs) // zufaelliger Eintrag aus probs (Eintraege von probs sind Gewichte)
{
   int stateidx;
   unsigned int i;
   double R, probsum;

   probsum = 0.0;
   for(i=0;i<probs.size();i++)
     probsum += probs[i];
     
   R = probsum*randomunitintervall();
   stateidx = -1;
   probsum = 0.0;
   do {
     stateidx++;
     if(stateidx>=(int)probs.size()) {
         cerr << "Error in funnction indel::randomchoice\n"
              << "variable stateidx out of range\n";
         exit(EXIT_FAILURE);
     }
     probsum+=probs[stateidx];
   } while(probsum<R-DBL_MIN);

   return stateidx;
}

double rand_exp(double x) {
    /* gibt exp(x)-verteilte ZVE zurueck.
     * (exp-verteilt mit Erwartungswert 1/x)
     */
    double R=randomunitintervall();
    if ( R>0.99 ) return -log(R)/x;
    else return -log(0.99)/x+rand_exp(x);
}
