/****************************************************************************
 * This file is part of the program kullerlein                              *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@eva.mpg.de                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <iostream>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include <ctime>
#include <string>
#include <cmath>
#include <cfloat>
#include <algorithm>

using namespace std;

#include "seqcode.hh"
#include "random.hh"
#include "substmod.hh"
#include "indel.hh"
#include "treespec.hh"


/*********************************************************************
 ************************  class clustal  ****************************
 *********************************************************************/


clustal::clustal(const char *filename)
{
   ifstream datei(filename);
   if(!datei) {
      cerr << "cannot open file " << filename << endl;
      exit(EXIT_FAILURE);
   }
   
   char c;
   int flag = 0;  // flag==1 => innerhalb von [], ==2 => ';'
   while(flag != 2 && datei.get(c)) {
     switch (c) {
        case ';' :
	  if(flag==0)
	  {
	    *this += c;
	    flag = 2;
	  }
	  break;
	case '[' :
	  flag = 1;
	  break;
	case ']' :
	  if(flag==1)
	    flag=0;
	  else {
	    cerr << "Unreadable tree format\n";
	    exit(EXIT_FAILURE);
	  }
	  break;
	default :
	  if(isprint(c) && !isspace(c) && flag==0)
	    *this += c;
     }
   }
   if(flag!=2) {
      cerr << "No semicolon at the end of the tree\n";
      exit(EXIT_FAILURE);
   }
   
   datei.close();
}

clustal::clustal(ifstream& datei)
{
   char c;
   int flag = 0;  // flag==1 => innerhalb von [], ==2 => ';'
   while(flag != 2 && datei.get(c)) {
     switch (c) {
        case ';' :
	  if(flag==0)
	  {
	    *this += c;
	    flag = 2;
	  }
	  break;
	case '[' :
	  flag = 1;
	  break;
	case ']' :
	  if(flag==1)
	    flag=0;
	  else {
	    cerr << "Unreadable tree format\n";
	    exit(EXIT_FAILURE);
	  }
	  break;
	default :
	  if(isprint(c) && !isspace(c) && flag==0)
	    *this += c;
     }
   }
   if(flag!=2) {
      cerr << "No semicolon at the end of the tree\n";
      exit(EXIT_FAILURE);
   }
}



/*********************************************************************
 ************************  class Newick  *****************************
 *********************************************************************/


Newick::Newick(string v) {
   if(v.size()==0) {
      cerr << "An empty string was passed to the constructor Newick\n";
      exit(EXIT_FAILURE);
   }
   desc_lab_len(v);
   subtrees();
}

void Newick::desc_lab_len(string v)
{

  string::size_type pos1, pos2, pos3;
     
  // finde branch_length und desc_list bzw. leaf_label;
  pos1 = v.find_last_of(')');
  pos2 = v.find_last_of(':');
  pos3 = v.find_last_of(';');
  if(pos3!=string::npos && (pos3<pos1 || pos3<pos2))
  {
    cerr << "Error in Clustal tree: \n"
         << v << "\n";
    exit(1);
  }
  if(pos1==string::npos) {
    leaf=1; // keine Klammern
    if(pos2!=string::npos)
    {
      if(pos2 > 0)
        desc_list = v.substr(0,pos2);
      else {
	cerr << "Leaf has no name\n";
	exit(1);
      }
    }
    else 
      desc_list = v;
  }
  else
  {
    leaf=0;
    desc_list = v.substr(0,pos1+1); 
  }
  if((pos2>pos1 || pos1==string::npos) && pos2!=string::npos)
  { 
    if(pos3!=string::npos)
      len_to_par = 0.0;  // verwerfe finale Astlaenge von "(...):x;"
    else {
      // extrahiere Astlaenge
      if(sscanf(v.substr(pos2+1).c_str(),"%lf",&len_to_par)!=1) {
         cerr << "Cannot scan from string\n"
	      << v.substr(pos2+1) << "\n";
         exit(1); 
      }
    }
  }
  else {  // keine Astlaenge angegeben
    if(pos3!=string::npos) // Root
      len_to_par = 0.0;
    else
      len_to_par = DEFAULTBRANCH; // willkuerliche Astlaenge
  }
     
}

void Newick::subtrees()
{
  // spalte descendant_list auf
  // falls man noch nicht bei einem
  // Blatt angekommen ist;
  // u. U. muss der Baum kuenstlich
  // gewurzelt werden
  vector<int> klam(desc_list.size());
  int lk = 0;
  int rk = 0;
  string::size_type pos1, pos2, pos3;
  for(unsigned int i=0; i<klam.size(); i++)
  { // bestimme Hierarchiestufen
    if(desc_list[i]=='(') lk++;
    if(desc_list[i]==')') rk++;
    klam[i]=lk-rk; 
  }
  if(klam.back()!=0) {
     cerr << klam.back() << " more '(' than ')' or v.v. in subtree\n"
	  << desc_list << "\n";
     exit(1);
  }
  vector<string::size_type> compos;  // enthaelt die Positionen von passender Kommas
  pos1 = desc_list.find_last_of(')');
  pos3 = 0;
  do{ 
      pos3=desc_list.find(',',pos3+1);
      if(pos3!=string::npos)
      {
	if(klam[pos3]==1) // tiefste Aufspaltungen
	  compos.push_back(pos3);
      }
  }while(pos3!=string::npos);
  if(compos.size()>0) {  // passende Kommata gefunden
     Kinder.push_back(desc_list.substr(1,compos[0]-1));
     for(unsigned int i=1;i<compos.size();i++)
        Kinder.push_back(desc_list.substr(compos[i-1]+1,compos[i]-compos[i-1]-1));
     Kinder.push_back(desc_list.substr(compos.back()+1,pos1-compos.back()-1));
  }
  else { 
    if(leaf==0) {  // bis jetzt kein Blatt, aber auch keine Aufspaltungen
       leaf = 1;
       pos2 = desc_list.find_last_of(':');
       if(pos2!=string::npos)
       {
         if(sscanf(desc_list.substr(pos2+1).c_str(),"%lf",&len_to_par)!=1) {
            cerr << "Cannot scan from string\n"
	         << desc_list.substr(pos2+1) << "\n";
	    exit(1); 
         }
       }
       string str = desc_list.substr(1,pos2-1);
       desc_list.swap(str);
    }
  }
}

/*********************************************************************
 ************************  class baum  *******************************
 *********************************************************************/

baum::baum(identifier ida)
{
   id = ida;
}

baum::baum(baum& source)
{
   unsigned int i;
   neighbor nb2, nb3;
   bool flag;
   
   id = source.id;  // identifier wird kopiert
   pro = source.pro; // Sequenz wird kopiert
   flag = FALSE;
   for(i=0;i<source.nb.size();i++) { // fuer jeden Nachbarn
       if(source.nb[i].nachbar==NULL) {
            cerr << "One tree node has a null pointer as neighbour\n";
	    exit(EXIT_FAILURE);
       }
       nb2 = source.nb[i];  // Nachbar wird kopiert
       if(!nb2.toroot) { // kein 'Vorfahr'
          nb2.nachbar = new baum(*(source.nb[i].nachbar)); // Rekursion
          nb.push_back(nb2);
	  
	  nb3 = source.nb[i].nachbar->nb[source.nb[i].nachbar->whichone(&source)];
	  nb.back().nachbar->nb.push_back(nb3);
	  nb.back().nachbar->nb.back().nachbar=this;
       } else {
          if(flag) {
	      cerr << "More than one parent in the\n"
		   << "copy constructor of class baum\n";
	      exit(EXIT_FAILURE);
	  }
	  flag = TRUE;
       }
   }
}

baum::baum(baum& source, bool topologyonly)
{  // kopiert nur die Topologie
   if(topologyonly) { // sollte immer der Fall sein
        unsigned int i;
        neighbor nb2, nb3;
        bool flag;
   
        id = source.id;  // identifier wird kopiert
        // Sequenz wird nicht kopiert
        flag = FALSE;
        for(i=0;i<source.nb.size();i++) { // fuer jeden Nachbarn
            if(source.nb[i].nachbar==NULL) {
                 cerr << "One tree node has a null pointer as neighbour\n";
	         exit(EXIT_FAILURE);
            }
            nb2 <= source.nb[i];  // Nachbar wird ohne ausrichter kopiert
            if(!nb2.toroot) { // kein 'Vorfahr'
               nb2.nachbar = new baum(*(source.nb[i].nachbar),topologyonly); // Rekursion
               nb.push_back(nb2);
	  
	       nb3 = source.nb[i].nachbar->nb[source.nb[i].nachbar->whichone(&source)];
	       nb.back().nachbar->nb.push_back(nb3);
	       nb.back().nachbar->nb.back().nachbar=this;
            } else {
               if(flag) {
	           cerr << "More than one parent in the\n"
		        << "copy constructor of class baum\n";
	           exit(EXIT_FAILURE);
	       }
	       flag = TRUE;
            }
        }
   } else {
        cerr << "Improper use of the topologyonly copy constructor of class baum\n";
	exit(EXIT_FAILURE);
   }
}

baum::baum(string s, bool anf)   // setzt identifier-Nummern auf -1
{
   // liest Namen und Astlaengen
   // aus einem String im Newick-Format
   Newick club(s);
   neighbor neuer;
   unsigned int i, j;
   bool flag;
   
   // Identifizierung
   if(!club.leaf) {
      id.number=-1;
      id.name="";
      id.treefromhere="";
   }
   else {
      id.number=-1;
      id.name=club.desc_list;
      id.treefromhere=id.name;
   }
   
   if(!anf) {
      // Richtung parent
      neuer.length=club.len_to_par*100.0;
      neuer.fix=TRUE;  // Aeste aus string s werden nicht veraendert
      nb.push_back(neuer);
   }
   
   // alle Kinder   
   for(i=0;i<club.Kinder.size();i++) {
      neuer.length=0.0;  // wird spaeter veraendert
      neuer.fix=TRUE;  // Aeste aus string s werden nicht veraendert
      neuer.toroot=FALSE;
      neuer.nachbar=new baum(club.Kinder[i],FALSE); // Rekursion
      nb.push_back(neuer);
      // Kinder andocken
      flag = FALSE;
      if(nb.back().nachbar!=NULL) {
         for(j=0;j<nb.back().nachbar->nb.size();j++) {  // NULL pointer zeigt Richtung parent
           if(nb.back().nachbar->nb[j].nachbar==NULL) {
	       if(flag==TRUE) {  // mehr als einen Vorfahren gibt's nicht
	          cerr << "More than one pointer was NULL in the\n"
		       << "copy constructor of class baum\n";
		  exit(EXIT_FAILURE);
	       }
	       flag = TRUE;
	       nb.back().nachbar->nb[j].nachbar=this;  // ein Nachbar des Kindes ist der Vorfahr
	       nb.back().fix=nb.back().nachbar->nb[j].fix;  // Kind bestimmt, ob Ast veraendert wird
	       nb.back().length=nb.back().nachbar->nb[j].length;  // Laenge anpassen
	       nb.back().nachbar->nb[j].toroot=TRUE;  // angedockt
	   }
         }
      }
   }
}


void baum::NeighborJoining(vector<vector<double> >& distmat, unsigned int& taxcount)
// NJ-Baum unter Beruecksichtigung der User-Wuensche bzgl. der Topologie
{
   if(!isLeaf() || isRoot()) {
      vector<int> children;   // identifier der Kinder
      for(unsigned int i=0;i<nb.size();++i)
         if(!nb[i].toroot) {
	    nb[i].nachbar->NeighborJoining(distmat,taxcount);
	    children.push_back(nb[i].nachbar->id.number);
      }
      bool flag=FALSE;
      if(children.size()>=2)
         flag=TRUE;
      while(flag) {
         double minmdist=DBL_MAX;
         int childA=-1, childB=-1, idxA=-1, idxB=-1;
         vector<double> netdivs(distmat.size());  // net divergences
         vector<vector<double> > mdists(distmat.size(),vector<double>(distmat.size())); // corrected distances

         for(unsigned int i=0;i<distmat.size();++i) {
             netdivs[i]=0.0;
	     for(unsigned int j=0;j<distmat.size();++j)
	         netdivs[i]+=distmat[i][j];      
         }
         for(unsigned int i=0;i<distmat.size();++i)	    
	     for(unsigned int j=0;j<distmat.size();++j)
	         mdists[i][j]=distmat[i][j]-(netdivs[i]+netdivs[j])/(double)(taxcount-2);
         for(unsigned int i=0;i<children.size();++i)
             for(unsigned int j=i+1;j<children.size();++j) {
	         if(mdists[children[i]][children[j]]<minmdist) {
	            minmdist=mdists[children[i]][children[j]];
		    childA=children[i]; childB=children[j];
	         }    
	     }
         if(childA==-1 || childB==-1) {
	    cerr <<"Error in the neighbor joining algorithm\n";
	    exit(EXIT_FAILURE);
	 }
	 for(unsigned int i=0;i<nb.size();++i)
	     if(!nb[i].toroot){
	        if(nb[i].nachbar->id.number==childA)
		   idxA=i;
		else
	           if(nb[i].nachbar->id.number==childB)
		      idxB=i;
	 }
         if(idxA==-1 || idxB==-1) {
	    cerr <<"Error in the neighbor joining algorithm\n";
	    exit(EXIT_FAILURE);
	 }
         int newid=distmat.size();
	 double disttoA=distmat[childA][childB]/2.0+(netdivs[childA]-netdivs[childB])/(2.0*(double)(taxcount-2)),
	        disttoB=distmat[childA][childB]-disttoA;
         if(disttoA<DBL_MIN) {
	    //cerr << "There are negative distances\n";
	    disttoA=0.1*distmat[childA][childB];
	    disttoB=distmat[childA][childB]-disttoA;
	 } else {
            if(disttoB<DBL_MIN) {
	       //cerr << "There are negative distances\n";
	       disttoB=0.1*distmat[childA][childB];
	       disttoA=distmat[childA][childB]-disttoB;
	    }
	 }
	 if(children.size()>2) {
            // neuen Knoten einfuegen
            vector<neighbor> newnb;
            baum *p = new baum(id);
	    p->id.number=newid;
	    for(int i=0;i<(int)nb.size();++i) {
	        if(i!=idxA && i!=idxB) {
	           newnb.push_back(nb[i]);
	        } else {
	           p->nb.push_back(nb[i]);
	           baum *t = p->nb.back().nachbar;
	           t->nb[t->whichone(this)].nachbar=p;
		   if(i==idxA)
		      p->nb.back().length=disttoA;
		   else
		      p->nb.back().length=disttoB;
		   t->nb[t->whichone(p)].length=p->nb.back().length;
	        }
	    }
            nb = newnb;
            neighbor neu;  // Konstruktor setzt fix auf FALSE;
            nb.push_back(neu);
            nb.back().nachbar=p;
            p->nb.push_back(neu);
            p->nb.back().toroot=TRUE;
            p->nb.back().nachbar=this;
         } else {
	     flag=FALSE;
	     id.number=newid;
	     nb[idxA].length=disttoA;
	     nb[idxA].nachbar->nb[nb[idxA].nachbar->whichone(this)].length=disttoA;
	     nb[idxB].length=disttoB;
	     nb[idxB].nachbar->nb[nb[idxB].nachbar->whichone(this)].length=disttoB;
	 }
	 // distmat veraendern
	 --taxcount;
	 for(int i=0;i<(int)distmat.size();++i) {
	     if(i!=childA && i!=childB) {
	           distmat[i].push_back((distmat[childA][i]+distmat[childB][i]-distmat[childA][childB])/2.0);
                   if(distmat[i][newid]<DBL_MIN) {
	              //cerr << "There are negative distances which are now set to "<<0.00001<<endl;
	              distmat[i][newid]=0.00001;
	           }
	     } else {
	        if(i==childA)
		   distmat[i].push_back(disttoA);
		else
		   distmat[i].push_back(disttoB);
	     }   
         }
	 distmat.push_back(vector<double>(newid+1));
	 for(int i=0;i<newid;++i)
             distmat[newid][i]=distmat[i][newid];
	 distmat[newid][newid]=0.0;
	 children.clear();  
         for(unsigned int i=0;i<nb.size();++i)
	     if(!nb[i].toroot)
	       children.push_back(nb[i].nachbar->id.number);
      }   
      if(isRoot()) {
         if(!isLeaf()) {
	    cerr << "The Neighbor Joining algorithm works only with a leaf as the root!\n";
	    exit(EXIT_FAILURE);
	 }
	 for(unsigned int i=0;i<nb.size();++i) {
	     nb[i].length=distmat[id.number][nb[i].nachbar->id.number];
	     nb[i].nachbar->nb[nb[i].nachbar->whichone(this)].length=nb[i].length;   
	 }
      }
   }
}




baum::~baum()
{
   for(unsigned int i=0;i<nb.size();i++) { // fuer jeden Nachbarn
      if(!nb[i].toroot) // kein 'Vorfahr'
         delete(nb[i].nachbar);
   }
}

baum *baum::NB(int i) const
{
   if(i<0 || i>=(int)nb.size()) {
       cerr << "Index out of range in baum::NB(int i)\n";
       exit(EXIT_FAILURE);
   }
   
   return nb[i].nachbar;
}

int  baum::whichone(const baum *T) const
{
   int idx;

   if(T==NULL) {
      cerr << "Argument of whichone(baum *T) is NULL\n";
      exit(EXIT_FAILURE);
   }
   
   idx=-1;
   for(unsigned int i=0;i<nb.size();i++) {
      if(nb[i].nachbar==T) {
         if(idx!=-1) {
	    cerr << "One node appears twice in baum::whichone(baum *T)\n";
	    exit(EXIT_FAILURE);
	 }
	 idx=(int)i;
      }
   }
   
   if(idx==-1) {
      cerr << "Node T could not be found in baum::whichone(baum *T)\n";
      exit(EXIT_FAILURE);
   }   
   
   return idx;
}

baum* baum::parent() const
{
   unsigned int i;
   baum *p;
   
   p=NULL;
   for(i=0;i<nb.size();i++) {
       if(nb[i].toroot) {
          if(p!=NULL) {
	     cerr << "There is more than one parent in baum::parent()\n";
	     exit(EXIT_FAILURE);
	  }
	  p=nb[i].nachbar;
       }
   }
   
   return p;
}

baum* baum::root()
{
   baum *p, *r;
   
   r=this;
   p = parent();
   if(p!=NULL)
        r=p->root();
   
   return r;
}

size_t baum::size() const
{
   size_t gr;
   gr=1;

   for(unsigned int i=0;i<nb.size();i++)
      if(!nb[i].toroot)
         gr+=nb[i].nachbar->size();
	    
   return gr;
}

size_t baum::numleaves() const
{ // gibt die Anzahl der Blaetter, die nicht in Richtung der Wurzel liegen, zurueck
   size_t gr;
   
   if(isLeaf())
     gr=1;  // keine Rekursion, falls schon Blatt
   else {
     gr=0;
     for(unsigned int i=0;i<nb.size();i++)
      if(!nb[i].toroot)
         gr+=nb[i].nachbar->numleaves();
   }
	    
   return gr;
}


double baum::laenge() const
{
    double lsum=0.0;
    
    for(unsigned int i=0;i<nb.size();++i)
        if(!nb[i].toroot) {
	   lsum+=nb[i].nachbar->laenge();
	   lsum+=nb[i].length;
        }
    
    return lsum;
}

double baum::aprxspan() const
{  // mittlerer 'Durchmesser' des Baumes
   double hsum = 0.0;
   int numchil = 0;
   
   for(unsigned int i=0;i<nb.size();++i)
       if(!nb[i].toroot) {
          hsum+=nb[i].nachbar->aprxspan();
	  hsum+=nb[i].length;
	  ++numchil;
       }
       
   if(numchil>0)
      return hsum/(double)numchil;
   else
      return 0.0;
}


double baum::aprxheight() const
{
  // ungefaehre Hoehe des Baumes (was immer das ist);
  // nur annaehernd sinnvoll, wenn die Wurzel ein Blatt ist
  return aprxspan();
}


double baum::opt_delta() const
{
  // einer Nebenbemerkung in Mau et al. 1999 zufolge
  // war zumindest in einem Beispiel das optimale Delta = 3% der Hoehe eines typischen Baumes
  // (sollte eigtl. im Burn-in angepasst werden)
  
  return 0.03*aprxheight();
}

baum *baum::findLeaf(identifier& id2)       // findet Blatt mit Wert id2
                                            // Man muss an der Wurzel starten!!!
{
   baum *p=NULL;
   
   if(isLeaf() && id==id2)
      p=this;
   else {
      unsigned int i=0;
      while(p==NULL && i<nb.size()) {
         if(!nb[i].toroot)
            p=nb[i].nachbar->findLeaf(id2);
	 i++;
      }   
   }
   
   return p;
}

size_t baum::depth() const
{
   baum *p;
   
   if(isRoot())
      return 0;
   else {
      p=parent();
      return 1+p->depth();
   }
}

/* so nicht sinnvoll
size_t baum::height() const
{
   int max, akt;
   
   if(isLeaf())
      return 0;
   else {
      max=0;
      for(unsigned int i=0;i<nb.size();i++) {
         if(!nb[i].toroot) {
            akt=nb[i].nachbar->height();
            if(max<akt)
               max=akt;
	 }
      }
      return 1+max;
   }   
}
*/

bool baum::isLeaf() const
{
   if(nb.size()<=1)
      return TRUE;
   else
      return FALSE;
}

bool baum::isRoot() const
{
   baum* p;
   
   p=parent();
   if(p==NULL)
      return TRUE;
   else
      return FALSE;
}
   

void baum::setid(const identifier& ida)
{
  id = ida;
}


void baum::swapbr(int nbkn, int ownch, int nbch) // tauscht Kind nbch von nbkn mit eigenem Kind ownch (Vorsicht mit Wurzel!!!!)
{
   baum *nbkind, *owkind, *nghbr;
   bool owpflag, nbpflag;  // Wo sitzen die Eltern vom Nachbarn und von this
   int owidx, nbidx;
   
   nghbr=NB(nbkn);
   owkind=NB(ownch);
   nbkind=nghbr->NB(nbch);
   owidx=owkind->whichone(this);
   nbidx=nbkind->whichone(nghbr);
   
   if(nghbr==owkind || nbkind==owkind || nbkind==this) {
      cerr << "There are loops in the tree\n";
      exit(EXIT_FAILURE);
   }

   if(owkind==parent())
      owpflag=TRUE;
   else
      owpflag=FALSE;
   if(nbkind==nghbr->parent())
      nbpflag=TRUE;
   else
      nbpflag=FALSE;
   if(nbpflag && owpflag) {
      cerr << "The tree seems to have more than one root!\n";
      exit(EXIT_FAILURE);
   }  
   
   nb[ownch].nachbar=nbkind;
   nb[ownch].length=nbkind->nb[nbidx].length;
   nb[ownch].fix=nbkind->nb[nbidx].fix;
   nbkind->nb[nbidx].nachbar=this;

   nghbr->nb[nbch].nachbar=owkind;
   nghbr->nb[nbch].length=owkind->nb[owidx].length;
   nghbr->nb[nbch].fix=owkind->nb[owidx].fix;
   owkind->nb[owidx].nachbar=nghbr;
   
   if(nb[nbkn].fix) {
      cerr << "Destroyed a branch which should have been fixed!\n";
      exit(EXIT_FAILURE);
   }

   if(nbpflag)
      nbkind->root()->reroot(NULL);
   if(owpflag)
      owkind->root()->reroot(NULL);
}


double baum::rndbrl(double oldl, double delta, bool fix)
{
   double R, r1, r2, ai, border=-DBL_MAX;
   
   if(fix)
      border=0.0;
   
   if(oldl-delta > border)  // ziehe aus Gleichverteilung
       R = 2.0*delta*randomunitintervall() - delta;
   else {  // ziehe aus Mixtur
       ai = oldl - border;
       r1 = (delta-ai)*randomunitintervall();
       r2 = 2.0*ai*randomunitintervall() + delta - ai;
       R = (r1*(delta-ai) + r2*ai)/delta - ai;
   }
   
   return oldl+R;
}


double baum::shiftnodes(double delta, int idx) //  verschiebt zwei Nachbarn (this und nb[idx].nachbar) ,
                                               // gibt neue Astlaenge zurueck
{
   double newl;
   
   // neue Astlaenge ziehen
   newl = rndbrl(nb[idx].length,delta,nb[idx].fix);
   
   nb[idx].length = abs(newl);
   nb[idx].nachbar->nb[nb[idx].nachbar->whichone(this)].length = abs(newl);
   
   // falls newl negativ ist, muessen noch Aeste vertauscht werden
   return newl;
}

double baum::shiftnodes(double delta, int idx, bool allfix) //  verschiebt zwei Nachbarn (this und nb[idx].nachbar) ,
                                                            // gibt neue Astlaenge zurueck
{
   if(allfix) {
      double newl;
      // neue Astlaenge ziehen
      newl = rndbrl(nb[idx].length,delta,TRUE);  // nie die Topologie aendern
      nb[idx].length = newl;
      nb[idx].nachbar->nb[nb[idx].nachbar->whichone(this)].length = newl;
      return newl;
   } else
      return shiftnodes(delta,idx);
}
/*
void baum::maunewton(double delta)
{
   vector<baum *> nodes; // nodes[i] und nbidx[i] kennzeichnen eine Kante
   vector<int> nbidx;    // nbidx[i] ist der Index eines Nachbarn von nodes[i]
   vector<int> zuford;   // zufaellige Ordnung der Knoten
   double newl;
   int ownch, nbch;
   baum *p;
   baum *q;

   if(!isRoot())
      reroot(this);
      
   edges(nodes,nbidx); // alle Kanten werden erfasst
   rndorder(zuford,nodes.size()); // zufaellige Ordnung
   
   for(unsigned int i=0;i<zuford.size();i++) {
       // Laenge der Kante zwischen nodes[zuford[i]] und nodes[zuford[i]]->nb[nbidx[zuford[i]]].nachbar aendern
       p=nodes[zuford[i]];
       q=p->NB(nbidx[zuford[i]]);
       newl = p->shiftnodes(delta,nbidx[zuford[i]]);
       if(newl<0.0) { // Topologie aendern
          // zufaellige Kinder von p und q, die nicht p oder q sind
          ownch=p->rndchild(nbidx[zuford[i]]);
	  nbch=q->rndchild(q->whichone(p));
          swapbr(nbidx[zuford[i]],ownch,nbch);
	  
// sampeln
	  
	  // Die Kante zwischen p und seinem ownch-ten Kind
	  // wird jetzt durch q und dessen nbch-ten Kind beschrieben
          keeptrack(nodes,nbidx,zuford[i],ownch,nbch);
       }
   }
}
*/

void baum::update_tree() // nur Teilbaum, falls man nicht mit der Wurzel beginnt
{
   char laenge[20];
   bool flag=FALSE;
   unsigned int i;

  if(!isLeaf() || isRoot()) {
      id.treefromhere="(";
   
      if(isRoot() && isLeaf())  // falls die Wurzel ein Blatt ist, muss am Schluss noch ihr Name angehaengt werden
         flag=TRUE;
      
      for(i=0;i<nb.size();i++) {  // fuer alle Nachbarn
         if(!nb[i].toroot) { // nicht im Kreis gehen
            nb[i].nachbar->update_tree();
	    if(flag==FALSE)
               sprintf(laenge,"%.6f",nb[i].length);
	    else
               sprintf(laenge,"%.6f",nb[i].length/2.0); // bei der Situation 'Wurzelblatt' wird die Laenge auf beide Cluster aufgeteilt
            id.treefromhere+=nb[i].nachbar->id.treefromhere; // Name des Clusters
	    id.treefromhere+=":";
	    id.treefromhere+=laenge; // Laenge des Astes
	    if(flag)  //nb.size()<2
	      id.treefromhere+=",";
	    else {
	      if(i<nb.size()-2 || (i==nb.size()-2 && !nb.back().toroot)) // bis zum vorletzten Kinderast 
	        id.treefromhere+=",";
	      else
	        id.treefromhere+=")";
	    }
         } 
       }
    
       if(flag==TRUE) {  // ist die Wurzel kein Blatt, ist das hier nicht notwendig
          id.treefromhere+=id.name;
          id.treefromhere+=":";
          id.treefromhere+=laenge;  // halbe Astlaenge, s.o.
          id.treefromhere+=")";
       }

       id.treefromhere+="\0"; // Abschluss
    } else
       id.treefromhere=id.name;
}

/* unnoetig
void baum::setfixes() // nur fuer Teilbaum, falls man nicht mit der Wurzel beginnt
{
   if(!isLeaf()) {
      id.name="";   

      for(unsigned int i=0;i<nb.size();i++) {  // fuer alle Nachbarn
         if(!nb[i].toroot) { // nicht im Kreis gehen
            nb[i].nachbar->setfixes();
            id.name+=nb[i].nachbar->id.name; // Name des Clusters
	    if(i<nb.size()-1 && !nb.back().toroot) // bis zum vorletzten Kinderast 
	       id.name+=",";
	    else
	       id.name+="\0";
         }
       }
       if(!isRoot()) {
          char usrin;
	  baum *p=parent();
          cout << "Keep OTU (" << id.name << ") fixed? (Y/N)\n";
          do {
            cin >> usrin;
          } while(toupper(usrin)!='Y' && toupper(usrin)!='N');
	  if(toupper(usrin)=='Y') {
	     p->nb[p->whichone(this)].fix=TRUE;
	     nb[whichone(p)].fix=TRUE;
	  } else {
	     p->nb[p->whichone(this)].fix=FALSE;
	     nb[whichone(p)].fix=FALSE;
	  }
       }
    }
}
*/

void baum::print(ostream& trstr)
{
   if(!isRoot())
      reroot(NULL); //this wird root
   
   update_tree();

   trstr << id.treefromhere<<";\n";
}

void baum::allebisauf(vector<int> &v, unsigned int size, int except)
{
   v.resize(size);
   
   for(int i=0;i<(int)size;i++)
      if(i==except)
         v[i]=0;
      else
         v[i]=1; 
}
   
   
baum *baum::pickrandleaf()  // waehlt zufaellig ein Blatt
{
   baum *p;
   vector<int> AnzBl;
   
   if(isLeaf() && !isRoot())
      return this;
   if(nb.size()==0)
      return this;
   
   for(unsigned int i=0;i<nb.size();i++) {
      if(nb[i].toroot)
         AnzBl.push_back(0);
      else
         AnzBl.push_back(nb[i].nachbar->numleaves());
   }
   if(isLeaf() && isRoot())  // falls Wurzel ein Blatt ist, ist AnzBl um einen Eintrag laenger als nb
      AnzBl.push_back(numleaves());
   
   int choice = randomchoice(AnzBl);
   if(choice >= (int)nb.size())
      p = this;
   else 
      p = nb[choice].nachbar->pickrandleaf();
   
   return p;
}

baum *baum::reroot()  // macht eins der Blaetter zur Wurzel
{
   baum *p;
   
   p = pickrandleaf();
   p->reroot(NULL);
   
   return p;
}

void baum::reroot(baum *newp)  // macht Wurzel zur Wurzel
{
    for(unsigned i=0;i<nb.size();i++) { 
       int idx = nb[i].nachbar->whichone(this);
       if(nb[i].nachbar!=newp) {
          nb[i].nachbar->nb[idx].toroot=TRUE;
	  nb[i].toroot=FALSE;
       }
       if(!nb[i].toroot)
          nb[i].nachbar->reroot(this); // Rekursion
    }
}


void baum::bifurcate(bool anf) // erzeugt vollstaendig bifurkierenden Baum ohne ueberfluessige Knoten
{
    if(anf==TRUE)
       reroot(NULL); //this wird root
    fusenodes();
    splitnode();
    for(unsigned int i=0;i<nb.size();i++) {
       if(!nb[i].toroot)
          nb[i].nachbar->bifurcate(FALSE);
    }   
}

void baum::fuseblindnodes(bool anf) // verhindert im ganzen Baum *-*-* (wie bifurcate ohne Aufruf von splitnode)
{
    if(anf==TRUE)
       reroot(NULL); //this wird root
    fusenodes();
    for(unsigned int i=0;i<nb.size();i++) {
       if(!nb[i].toroot) {
         if(nb[i].nachbar!=NULL)
          nb[i].nachbar->fuseblindnodes(FALSE);
	 else {
	    cerr << "Neighbor points to nil\n";
	    exit(EXIT_FAILURE);
	 }
       }
    }   
}


void baum::splitnode() // loest anfaengliche Multifurkationen auf
{
    if(nb.size()>3) {
       unsigned int i;
       vector<bool> lr(nb.size());
       bool flag1=TRUE, flag2=TRUE;
       for(i=0;i<lr.size();++i) {
           if(nb[i].toroot)
	      lr[i]=FALSE;
	   else {
	      if(randomunitintervall()<0.5)
	         lr[i]=TRUE;
	      else 
	         lr[i]=FALSE;
	      flag1=(flag1 && lr[i]);
	      flag2=(flag2 && !lr[i]);
	   }
       }
       if(flag1) { // alle Eintraege von lr TRUE
          if(nb[0].toroot)
	     lr[1]=FALSE;
	  else
	     lr[0]=FALSE;  
       }
       if(flag2) { // alle Eintraege von lr FALSE
          if(nb[0].toroot)
	     lr[1]=TRUE;
	  else
	     lr[0]=TRUE;  
       }
       // Kopieren
       vector<neighbor> newnb;
       baum *p = new baum(id);
       baum *q = new baum(id);
       for(i=0;i<nb.size();i++) {
           if(lr[i]) { // nach p schieben
	        p->nb.push_back(nb[i]);
	        baum *t = p->nb.back().nachbar;
	        t->nb[t->whichone(this)].nachbar=p;
	   } else
	        if(nb[i].toroot)
	           newnb.push_back(nb[i]);
	        else {
	           q->nb.push_back(nb[i]);
	           baum *t = q->nb.back().nachbar;
	           t->nb[t->whichone(this)].nachbar=q;
		}
       }

       nb = newnb;
       neighbor neu;  // Konstruktor setzt fix auf FALSE;
       if(p->nb.size()>1) {
          nb.push_back(neu);
          nb.back().nachbar=p;
          p->nb.push_back(neu);
          p->nb.back().toroot=TRUE;
          p->nb.back().nachbar=this;
       
          // dem Ast zwischen this und p eine Astlaenge zuweisen,
          // die von den anderen Nachbarn von p abgezogen wird
          double minlength=DBL_MAX;
          for(i=0;i<p->nb.size();++i)
             if(p->nb[i].nachbar!=this && p->nb[i].length<minlength)
	        minlength=p->nb[i].length;
          minlength/=2.0;
          for(i=0;i<p->nb.size();++i) {
              if(p->nb[i].nachbar==this) {
	         p->nb[i].length=minlength;
	         nb[whichone(p)].length=minlength;
	      } else {
	         p->nb[i].length-=minlength;
	         baum *t = p->nb[i].nachbar;
	         t->nb[t->whichone(p)].length=p->nb[i].length;
	      }
          }
       } else {
          neu=p->nb[0];
	  baum *t = neu.nachbar;
	  t->nb[t->whichone(p)].nachbar=this;
	  p->nb.pop_back();
	  nb.push_back(neu);
	  delete(p);
       }
       if(q->nb.size()>1) {
          nb.push_back(neu);
          nb.back().nachbar=q;
          q->nb.push_back(neu);
          q->nb.back().toroot=TRUE;
          q->nb.back().nachbar=this;
          double minlength=DBL_MAX;
          for(i=0;i<q->nb.size();++i)
             if(q->nb[i].nachbar!=this && q->nb[i].length<minlength)
	        minlength=q->nb[i].length;
          minlength/=2.0;
          for(i=0;i<q->nb.size();++i) {
              if(q->nb[i].nachbar==this) {
	         q->nb[i].length=minlength;
	         nb[whichone(q)].length=minlength;
	      } else {
	         q->nb[i].length-=minlength;
	         baum *t = q->nb[i].nachbar;
	         t->nb[t->whichone(q)].length=q->nb[i].length;
	      }
          }
       } else {
          neu=q->nb[0];
	  baum *t = neu.nachbar;
	  t->nb[t->whichone(q)].nachbar=this;
	  q->nb.pop_back();
	  nb.push_back(neu);
	  delete(q);
       }
    }
}

void baum::fusenodes() // verhindert  *--*--*
{
    if(nb.size()==2) {
       // damit nicht der Parent weggeworfen wird:
       vector<neighbor> nbbuff(2);
       if(nb[0].toroot) {
         if(nb[1].toroot) {
	    cerr << "Impossible: two neighbours are parents\n";
	    exit(EXIT_FAILURE);
	 }
	 nbbuff=nb;
       } else {
         if(nb[1].toroot) {
	    nbbuff[0]=nb[1];
	    nbbuff[1]=nb[0];
	 } else
	    nbbuff=nb;
       }
       
       int leftidx = nbbuff[0].nachbar->whichone(this);
       // Laengen addieren
       nbbuff[0].length+=nbbuff[1].length;
       nbbuff[0].nachbar->nb[leftidx].length=nbbuff[0].length;
       // fix anpassen
       nbbuff[0].fix=(nbbuff[0].fix || nbbuff[1].fix);
       nbbuff[0].nachbar->nb[leftidx].fix=nbbuff[0].fix;
       // rechten Knoten eliminieren
       baum *p = nbbuff[1].nachbar;
       nbbuff.pop_back();
       id=p->id; // identifier kopieren (wichtig, falls p ein Blatt ist)
       pro=p->pro;
       for(unsigned int i=0;i<p->nb.size();i++) {
          if(p->nb[i].nachbar!=this) {
	     nbbuff.push_back(p->nb[i]);
	     nbbuff.back().nachbar->nb[nbbuff.back().nachbar->whichone(p)].nachbar=this;
	  }
	  p->nb[i].nachbar=NULL;
       }
       delete(p);
       nb=nbbuff;
    }
}

string baum::clustname()
{
   string s;
   vector<string> vs;
   unsigned int i;
   
   if(!isLeaf() || isRoot()) {
      for(i=0;i<nb.size();i++)
         if(!nb[i].toroot)
	    vs.push_back(nb[i].nachbar->clustname());
      sort(vs.begin(),vs.end());
      for(i=0;i<vs.size();i++) {
         s+=vs[i];
      }
      s+='.';
   }
   else
      s+=id.name;

   return s; 
}

int baum::rndchild(int ausser) // zufaelliger Nachbar!=nb[ausser]
{
   vector<int> waehlbar(nb.size());
   
   if(ausser <0 || ausser>=(int)nb.size())
   {
      cerr << "Argument to function rndchild out of range\n";
      exit(EXIT_FAILURE);
   }
   
   for(unsigned int i=0;i<waehlbar.size();i++) {
      if((int)i==ausser)
         waehlbar[i]=0;
      else
         waehlbar[i]=1;
   }
   
   int retval = randomchoice(waehlbar);
   
   return retval;
}

void baum::edges(vector<baum *>& nodes, vector<unsigned int>& nbidx)
{
    if(nodes.size()==0 && !isRoot()) // Anfang, aber nicht in der Wurzel
       reroot(this);

    for(unsigned int i=0;i<nb.size();i++) {
        if(!nb[i].toroot) {  // jede Kante, die nicht zur Wurzel fuehrt wird erfasst
	    nodes.push_back(this);
	    nbidx.push_back(i);
	    nb[i].nachbar->edges(nodes,nbidx);  
	}
    }
}

void baum::internaledges(vector<baum *>& nodes, vector<unsigned int>& nbidx)
{
    for(unsigned int i=0;i<nb.size();i++) {
        if(!nb[i].toroot) {  
	// jede innere Kante, die nicht zur Wurzel fuehrt wird erfasst
	    if(!isLeaf() && !nb[i].nachbar->isLeaf()) {
	       nodes.push_back(this);
	       nbidx.push_back(i);
	    }
	    nb[i].nachbar->internaledges(nodes,nbidx);  
	}
    }
}


void baum::intnodes(vector<baum *>& nodes)
{
    for(unsigned int i=0;i<nb.size();i++) {
        if(!nb[i].toroot) {
	    if(!isLeaf())  
	       nodes.push_back(this);
	    nb[i].nachbar->intnodes(nodes);  
	}
    }
}

/*******************   Funktionen zum extrahieren eines Teilbaums    *****************/

baum* baum::teilbaum(vector<string>& IDs, vector<char>& pattern)
// erzeugt Teilbaum, der Taxa, die im pattern durch ein Nukleotid vertreten sind
{
   baum *kopie=new baum(*this,TRUE); // kopiert ohne Sequenzen und ausrichter
   baum *retbaum = kopie->abuse_fix(IDs,pattern);  // bei Aesten des zu extrahierenden Teilbaums wird fix auf TRUE gesetzt
   retbaum->trim(); // schneidet alle Aeste mit fix=FALSE ab
   retbaum->fuseblindnodes(TRUE);  // verhindert *-*-*
   return retbaum;
}

baum* baum::abuse_fix(vector<string>& IDs, vector<char>& pattern)
{
   baum *p=NULL;
   vector<bool> nucpres;   // TRUE=>Nukleotid, FALSE=>GAP
   vector<bool> namefound; // wurden die Namen in IDs im Baum gefunden
   unsigned int i;
   bool flag = TRUE;
   
   // Initialisierung von nucpres und namefound
   if(IDs.size()!=pattern.size()) {
      cerr << "Unequal numbers of identifiers and of sequences in abuse_fix\n"
           << IDs.size()<<"\t"<<pattern.size()<<endl;
      exit(EXIT_FAILURE);
   }
   nucpres.resize(pattern.size());
   namefound.resize(pattern.size());
   for(i=0;i<pattern.size();++i) {
       namefound[i]=FALSE;
       if(pattern[i]=='_' || pattern[i]=='-')
          nucpres[i]=FALSE;
       else
          nucpres[i]=TRUE;
   }


   // Werte von fix setzen
   findnames(p,nucpres,IDs,namefound);
   // nur zur Kontrolle
   for(i=0;i<namefound.size();++i)
       flag = (flag && namefound[i]);
   if(!flag) {
      cerr << "At least one of the names was not found in the tree\n";
      exit(EXIT_FAILURE);
   }
   if(p==NULL) {
      cerr << "Lost tree\n";
      exit(EXIT_FAILURE);
   }
   // p wird Wurzel
   p->reroot(NULL);
   
   if(!p->post_order_abuse_fix(nucpres)) {
      cerr<<"Error in function abuse_fix\n";
      exit(EXIT_FAILURE);
   }
   p->pre_order_abuse_fix();
      
   return p;
}



void baum::findnames(baum*& p, vector<bool>& nucpres, vector<string>& IDs, vector<bool>& namefound)
{
   unsigned int i;
   
   for(i=0;i<nb.size();++i)
     if(!nb[i].toroot)
        nb[i].nachbar->findnames(p,nucpres,IDs,namefound);
	
   if(isLeaf()) {
      bool nameflag = TRUE;
      for(i=0;i<IDs.size() && nameflag;++i)
          if(id.name==IDs[i]) {
	     if(p==NULL && nucpres[i])
	        p=this;
	     nameflag=FALSE;
	     id.number=i;
	     if(namefound[i]) {
	        cerr << "One name was found twice in the tree\n";
		exit(EXIT_FAILURE);
	     }
	     namefound[i]=TRUE;
	  }
   }
}


bool baum::post_order_abuse_fix(vector<bool>& nucpres)
{
   unsigned int i;
   int rootidx = -1;
   bool fixflag = FALSE, fix2;
   
   for(i=0;i<nb.size();++i)
     if(!nb[i].toroot) {
        fix2 = nb[i].nachbar->post_order_abuse_fix(nucpres);
        fixflag = (fixflag || fix2);
     } else
        rootidx = (int)i;
	
   if(isLeaf())
      fixflag = nucpres[id.number];
   
   if(rootidx>-1) {
      nb[rootidx].fix=fixflag;
      nb[rootidx].nachbar->nb[nb[rootidx].nachbar->whichone(this)].fix=fixflag;
   }
   
   return fixflag;
}


void baum::pre_order_abuse_fix()
{
    int numfix = 0;
    unsigned int i;
    
    if(!isLeaf()) { 
      for(i=0;i<nb.size();++i)
         if(nb[i].fix)
	    ++numfix;
      if(numfix==1) { // zuruecksetzen
         for(i=0;i<nb.size();++i) {
	     nb[i].fix=FALSE;
	     nb[i].nachbar->nb[nb[i].nachbar->whichone(this)].fix=FALSE;
	 }
      }
    }
    
    for(i=0;i<nb.size();++i)
       if(!nb[i].toroot)
          nb[i].nachbar->pre_order_abuse_fix();
}


void baum::trim()
{// schneidet alle Aeste mit fix=FALSE ab
   vector<neighbor> nbnew;
   
   for(unsigned int i=0;i<nb.size();++i) {
       if(!nb[i].toroot) {
          if(!nb[i].fix)
	     delete(nb[i].nachbar);
	  else {
	     nb[i].nachbar->trim();
	     nbnew.push_back(nb[i]);
	  }
       } 
       else
         nbnew.push_back(nb[i]);
   }
   
   nb=nbnew;
}




bool baum::operator== (baum& b2)  // Vergleichsoperator fuer die Topologie
{
   bool retval;
   
   // alte Wurzeln merken
   baum *r1=root();
   baum *r2=b2.root();
   
   // zwei gleiche Blaetter suchen
   baum *l1=r1->pickrandleaf();
   baum *l2=r2->findLeaf(l1->id);
 
   // kein Blatt, das l1 entspricht
   if(l2==NULL)
      return FALSE;

   // umwurzeln   
   l1->reroot(NULL); //l1 wird root
   l2->reroot(NULL); //l2 wird root
   
   retval = (l1->clustname()==l2->clustname());
      
   // zurueckwurzeln
   r1->reroot(NULL); //r1 wird root
   r2->reroot(NULL); //r2 wird root
   
   return retval;
}

bool baum::evolve(substmodel& sub, indel& idmod)
{
   bool retval=TRUE;
   if(isRoot()) { // zufaellige Sequenz erzeugen
     // unsigned int slen=idmod.randlen();
      unsigned int slen=idmod.explen();
      for(unsigned int i=0;i<slen;++i)
         pro.push_back(sub.choosebase());
   }
   for(unsigned int i=0;i<nb.size();++i)
     if(!nb[i].toroot) {
         // Nachbarsequenz mit Alignment
	 sub.settime(nb[i].length);
	 idmod.fillP(nb[i].length);
	 int anc=0;
	 int thisidx=nb[i].nachbar->whichone(this);
	 zustand z=START;
	 while(z!=ENDE) {
	    z=idmod.nextpair(z,(anc<pro.laen()));
	    switch(z) {
	        case PAIR:
		   nb[i].nachbar->pro.push_back(sub.evolbase(pro[anc]));
		   nb[i].ausrichter.push_back(nb[i].nachbar->pro.laen()-1);
		   nb[i].nachbar->nb[thisidx].ausrichter.push_back(anc);
		   ++anc;
		   break;
		case INSERTION:
		   nb[i].nachbar->pro.push_back(sub.choosebase());
		   nb[i].nachbar->nb[thisidx].ausrichter.push_back(anc-1);
		   break;
		case DELETION:
		   nb[i].ausrichter.push_back(nb[i].nachbar->pro.laen()-1);
		   ++anc;
		   break;
		case ENDE:
		   if(anc!=pro.laen()) {
		      cerr << "Error in function baum::evolve\n";
		      exit(EXIT_FAILURE);
		   }
		   break;
		default:
	           cerr << "Impossible state in function baum::evolve\n";
		   exit(EXIT_FAILURE);
	    }
	 }
         // Rekursion
         retval=(retval && nb[i].nachbar->evolve(sub,idmod));
     }
     
     if(isLeaf() && pro.laen()==0)  // keine Sequenz an den Blaettern
        retval=FALSE;
	
     return retval;
}


void baum::clear()
{
   pro.clear();
   for(unsigned int i=0;i<nb.size();++i){
      nb[i].ausrichter.clear();
      if(!nb[i].toroot)
         nb[i].nachbar->clear();
   }
}

/*********************************************************************
 ************************  class identifier  *************************
 *********************************************************************/

void identifier::setnumber(vector<vector<char> > Namen)
{
   bool flag=FALSE;
   string s;
   unsigned int i,j;
   
   number=-1;
   for(i=0;i<Namen.size();i++) {
       s = "";
       for(j=0;j<Namen[i].size();j++)
           s+=Namen[i][j];
       if(name==s) {
          if(flag) {
            cerr << "Two sequences in the data set have the same name!\n";
	    exit(EXIT_FAILURE);
          } else {
	    flag=TRUE;
	    number=i;
	  }
       }
   }
   if(number==-1) {
      cerr << "Taxon " << name << " was not found in the following data set:\n";
      for(i=0;i<Namen.size();++i) {
          for(j=0;j<Namen[i].size();++j)
              cerr << Namen[i][j];
	  cerr << endl;
      }
      exit(EXIT_FAILURE); 
   }
}

identifier identifier::operator= (identifier id2)  // Zuweisungsoperator
{
   name=id2.name;
   number=id2.number;
   treefromhere=id2.treefromhere;

   return *this;
}

bool identifier::operator== (identifier& id2)  // Vergleichsoperator
{
   bool flag=TRUE;
   
   if(name!=id2.name)
      flag=FALSE;
      
   // number und treefromhere werden nicht auf Gleichheit ueberprueft
   
   return flag;
}

/*********************************************************************
 ************************  class neighbor  ***************************
 *********************************************************************/

neighbor::neighbor()
{
   length = 0.0;
   fix = FALSE;
   toroot = FALSE;
   nachbar = NULL;
}

neighbor neighbor::operator= (neighbor nb2)  // Zuweisungsoperator
{
   length = nb2.length;
   fix = nb2.fix;
   toroot = nb2.toroot;
   nachbar = nb2.nachbar;
   ausrichter = nb2.ausrichter;
   
   return *this;
}

neighbor neighbor::operator<= (neighbor nb2)  // Zuweisungsoperator(!!!) ohne Zuweisung vom ausrichter
{
   length = nb2.length;
   fix = nb2.fix;  
   toroot = nb2.toroot;
   nachbar = nb2.nachbar;
   
   return *this;
}


/***********************************************************************************
 ****************** class bioseq         *******************************************
 ***********************************************************************************/


bioseq::bioseq()
{}

bioseq::bioseq(vector<char>& seq, int typ)
{
   unsigned int i;
   char c;
   vector<int> states;

   for(i=0;i<seq.size();i++) {
      c = seq[i];
      if(c!='_' && c!='-') {
         switch(typ) {
             case DNA :
	       push_back(chartobase(c));
	       break;
             case PROTEIN :
	       push_back(chartoamac(c));
	       break;
	     default :
	       cerr << "Unknown data typ in function bioseq(vector<char>& seq, alphsize typ)\n";
	       exit(EXIT_FAILURE);
         }
      }
   }
}

bioseq::~bioseq()
{}

int bioseq::laen() const
{
    return (int)size();
}

